import argparse
import os
import sys
from pathlib import Path

from loguru import logger

from syftbox import __version__
from syftbox.app.manager import list_app
from syftbox.app.manager import main as app_manager_main
from syftbox.client.client import main as client_main
from syftbox.server.server import main as server_main


def print_debug():
    try:
        import os
        import platform
        import shutil

        import psutil
        import yaml

        from syftbox.lib import DEFAULT_CONFIG_PATH, ClientConfig

        config_path = os.environ.get("SYFTBOX_CLIENT_CONFIG_PATH", DEFAULT_CONFIG_PATH)
        client_config = None
        apps = []
        try:
            client_config = ClientConfig.load(config_path)
            apps = list_app(client_config, silent=True)
            client_config = client_config.to_dict()
        except Exception:
            pass

        syftbox_path = shutil.which("syftbox")

        debug_info = {
            "system": {
                "resources": {
                    "cpus": psutil.cpu_count(logical=True),
                    "architecture": platform.machine(),
                    "ram": f"{psutil.virtual_memory().total / (1024**3):.2f} GB",
                },
                "operating_system": {
                    "name": "macOS"
                    if platform.system() == "Darwin"
                    else platform.system(),
                    "version": platform.release(),
                },
                "python": {
                    "version": platform.python_version(),
                    "binary_location": sys.executable,
                },
            },
            "syftbox": {
                "command": syftbox_path or "syftbox executable not found in PATH",
                "apps": apps,
                "client_config_path": config_path,
                "client_config": client_config,
            },
            "syftbox_env": {
                key: value
                for key, value in os.environ.items()
                if key.startswith("SYFT")
            },
        }
        logger.info(yaml.dump(debug_info, default_flow_style=False))
    except Exception as e:
        logger.info(e)


def main():
    parser = argparse.ArgumentParser(description="Syftbox CLI")
    subparsers = parser.add_subparsers(
        dest="command",
        description="Valid syftbox commands",
        help="subcommand to run",
    )

    # Define the client command
    subparsers.add_parser("client", help="Run the Syftbox client")

    # Define the server command
    subparsers.add_parser("server", help="Run the Syftbox server")

    # Define the install
    app_parser = subparsers.add_parser(
        "app", help="Manage SyftBox apps.", description="Manages SyftBox Apps"
    )

    app_parser = subparsers.add_parser(
        "version", help="Show SyftBox version", description="Shows the version"
    )

    app_parser = subparsers.add_parser(
        "debug", help="Show SyftBox debug info", description="Shows the debug info"
    )

    app_parser = subparsers.add_parser(
        "path", help="Get Syftbox Import Path", description="Prints the python path"
    )

    args, remaining_args = parser.parse_known_args()

    if args.command == "client":
        # Modify sys.argv to exclude the subcommand
        sys.argv = [sys.argv[0]] + remaining_args
        client_main()
    elif args.command == "server":
        # Modify sys.argv to exclude the subcommand
        sys.argv = [sys.argv[0]] + remaining_args
        server_main()
    elif args.command == "app":
        sys.argv = [sys.argv[0]] + remaining_args
        app_manager_main(app_parser, remaining_args)
    elif args.command == "version":
        print(__version__)
    elif args.command == "debug":
        print_debug()
    elif args.command == "path":
        current_dir = Path(__file__).parent.parent
        print(os.path.abspath(current_dir))
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



import json
import os
import shutil
from types import SimpleNamespace

import pytest

from syftbox.app.install import (
    check_os_compatibility,
    clone_repository,
    load_config,
    sanitize_git_path,
)


def test_valid_git_path():
    path = "Example/Repository"
    output_path = sanitize_git_path(path)
    assert path == output_path


def test_valid_git_url():
    path = "Example/Repository"
    http_url = f"http://github.com/{path}"
    output_path = sanitize_git_path(http_url)
    assert path == output_path

    https_url = f"https://github.com/{path}"
    output_path = sanitize_git_path(https_url)
    assert path == output_path


def test_invalid_git_path():
    path = "..Example/../Repository"
    with pytest.raises(ValueError) as excpt:
        _ = sanitize_git_path(path)
        assert excpt.value == "Invalid Git repository path format."


def test_second_invalid_git_path():
    path = "http://example.com"
    with pytest.raises(ValueError) as excpt:
        _ = sanitize_git_path(path)
        assert excpt.value == "Invalid Git repository path format."


def test_clone_valid_repository():
    path = "OpenMined/logged_in"
    temp_path = clone_repository(path)
    assert os.path.exists(temp_path)
    shutil.rmtree(temp_path)


def test_clone_repository_to_an_existent_path():
    # First call will make the repository path exist
    path = "OpenMined/logged_in"
    temp_path = clone_repository(path)
    assert os.path.exists(temp_path)

    # Second call must clone it again without any exception (replaces the old one).
    temp_path = clone_repository(path)
    shutil.rmtree(temp_path)


def test_clone_invalid_repository():
    path = "InvalidUser/InvalidRepo"
    with pytest.raises(ValueError) as excpt:
        _ = clone_repository(path)
        assert (
            excpt.value
            == "The provided repository path doesn't seems to be accessible. Please check it out."
        )


def test_load_app_config():
    valid_json_config = {
        "version": "0.1.0",
        "app": {
            "version": "0.1.0",
            "run": {"command": ["python", "main.py"], "interval": "10"},
            "env": {},
            "platforms": ["linux"],
            "pre_install": ["pip", "install", "psutil"],
            "post_install": [],
            "pre_update": [],
            "post_update": [],
        },
    }
    with open("app.json", "w") as app_json_file:
        json.dump(valid_json_config, app_json_file, indent=4)

    app_config = load_config("app.json")
    assert app_config.version == valid_json_config["version"]
    assert app_config.app.version == valid_json_config["app"]["version"]
    assert app_config.app.run.command == valid_json_config["app"]["run"]["command"]
    assert vars(app_config.app.env) == valid_json_config["app"]["env"]
    assert app_config.app.platforms == valid_json_config["app"]["platforms"]
    assert app_config.app.pre_install == valid_json_config["app"]["pre_install"]
    assert app_config.app.pre_update == valid_json_config["app"]["pre_update"]
    assert app_config.app.post_update == valid_json_config["app"]["post_update"]
    os.remove("app.json")


def test_load_invalid_app_config():
    with open("app.json", "w") as app_json_file:
        json.dump("\nHello World: \n Test", app_json_file, indent=4)

    with pytest.raises(ValueError) as expt:
        load_config("app.json")
        assert expt.value == "File isn't in JSON format"

    os.remove("app.json")


def test_load_inexistent_app_config():
    with pytest.raises(ValueError) as expt:
        load_config("inexistent_app.json")
        assert expt.value == "Couln't find the json config file for this path."


def test_os_compatibility_compatible():
    app_config_mock = SimpleNamespace(
        **{
            "app": SimpleNamespace(
                **{
                    "platforms": ["darwin", "linux"],
                }
            ),
        }
    )

    check_os_compatibility(app_config_mock)


def test_os_compatibility_incompatible():
    app_config_mock = SimpleNamespace(
        **{
            "app": SimpleNamespace(
                **{
                    "platforms": ["different_os"],
                }
            ),
        }
    )
    with pytest.raises(OSError) as e:
        check_os_compatibility(app_config_mock)
        assert e.value == "Your OS isn't supported by this app."


def test_os_compatibility_without_config():
    app_config_mock = SimpleNamespace(**{"app": {}})

    check_os_compatibility(app_config_mock)



import datetime
import sys
from platform import platform

import requests
from pydantic import BaseModel, Field

import syftbox
from syftbox.lib.lib import ClientConfig
from syftbox.__version__ import __version__


class ErrorReport(BaseModel):
    client_config: dict
    server_syftbox_version: str | None = None
    client_syftbox_version: str = __version__
    python_version: str = sys.version
    platform: str = platform()
    timestamp: datetime.datetime = Field(
        default_factory=lambda: datetime.datetime.now(datetime.timezone.utc)
    )

    @classmethod
    def from_client_config(cls, client_config: ClientConfig):
        client_config.token = None
        return cls(
            client_config=client_config.to_dict(),
            server_version=try_get_server_version(client_config.server_url),
        )


def make_error_report(client_config: ClientConfig):
    return ErrorReport.from_client_config(client_config)


def try_get_server_version(server_url):
    try:
        # do not use the server_client here, as it may not be in bad state
        return requests.get(f"{server_url}/info").json()["version"]
    except Exception:
        return None



import argparse
import os
import sys
from collections import namedtuple
from pathlib import Path

from loguru import logger

from ..lib import DEFAULT_CONFIG_PATH, ClientConfig
from .install import install

config_path = os.environ.get("SYFTBOX_CLIENT_CONFIG_PATH", None)


def list_app(client_config: ClientConfig, silent: bool = False) -> list[str]:
    apps_path = Path(client_config.sync_folder + "/" + "apps")
    apps = []
    if os.path.exists(apps_path):
        files_and_folders = os.listdir(apps_path)
        apps = [app for app in files_and_folders if os.path.isdir(apps_path / app)]

    if len(apps):
        if not silent:
            logger.info("\nInstalled apps:")
            for app in apps:
                logger.info(f"✅ {app}")
    else:
        if not silent:
            logger.info(
                "\nYou have no apps installed.\n\n"
                f"Try:\nsyftbox app install OpenMined/github_app_updater\n\nor copy an app to: {apps_path}"
            )
    return apps


def uninstall_app(client_config: ClientConfig) -> None:
    logger.info("Uninstalling Apps")


def update_app(client_config: ClientConfig) -> None:
    logger.info("Updating Apps")


def upgrade_app(client_config: ClientConfig) -> None:
    logger.info("Upgrading Apps")


Commands = namedtuple("Commands", ["description", "execute"])


def make_commands() -> dict[str, Commands]:
    return {
        "list": Commands(
            "List all currently installed apps in your syftbox.", list_app
        ),
        "install": Commands("Install a new app in your syftbox.", install),
        "uninstall": Commands("Uninstall a certain app.", uninstall_app),
        "update": Commands("Check for app updates.", update_app),
        "upgrade": Commands("Upgrade an app.", upgrade_app),
    }


class CustomHelpFormatter(argparse.HelpFormatter):
    def add_arguments(self, actions):
        for action in actions:
            if action.dest == "command":
                commands = make_commands()
                action.choices = commands.keys()
                # Build help message with descriptions
                action.help = "\n".join(
                    [f"{cmd}: {commands[cmd].description}" for cmd in commands]
                )
        super().add_arguments(actions)


# Parsing arguments and initializing shared state
def parse_args():
    parser = argparse.ArgumentParser(
        description="Run the web application with plugins.",
    )

    commands = make_commands()

    # Add a subparser to the "app" parser to handle different actions
    parser.add_argument(
        "command", choices=commands.keys(), help="The command to execute"
    )

    parser.add_argument(
        "--config_path", type=str, default=DEFAULT_CONFIG_PATH, help="config path"
    )
    args, remaining_args = parser.parse_known_args()
    return args, remaining_args


def main(parser, args_list) -> None:
    # Color codes for Error prompts
    RED = "\033[38;5;210m"
    RESET = "\033[0m"
    YELLOW_PASTEL = "\033[38;5;229m"
    args, remaining_args = parse_args()
    try:
        client_config = ClientConfig.load(args.config_path)
    except Exception:
        print(
            f"\n{RED}Error:{RESET} Couldn't find the proper client_config.json in: {YELLOW_PASTEL}{args.config_path}{RESET}.\n\n"
            + "Please ensure that:\n"
            + "\t- The configuration file exists at the specified path.\n"
            + f"\t- Provide the proper path by adding the {YELLOW_PASTEL}--config_path /path/to/client_config.json{RESET}"
        )
        return
    commands = make_commands()
    # Handle the subcommands as needed
    if args.command:
        command = commands[args.command]
        sys.argv = [sys.argv[0]] + remaining_args
        result = command.execute(client_config)
        if result is not None:
            # we should make this a type
            if isinstance(result, tuple):
                step, exception = result
                logger.info(f"Error during {step}: ", str(exception))
    else:
        parser.print_help()



import json
import os
from datetime import datetime

from syftbox.lib import ClientConfig


def main():
    # Load the client configuration
    config_path = os.environ.get("SYFTBOX_CLIENT_CONFIG_PATH", None)
    client_config = ClientConfig.load(config_path)

    # Get the current timestamp
    current_timestamp = datetime.now().isoformat()

    # Prepare the data to be written
    timestamp_data = {"last_check_in": current_timestamp}

    # Prepare output folders
    output_folder = f"{client_config.sync_folder}/{client_config.email}/app_pipelines/timestamp_recorder/"
    os.makedirs(output_folder, exist_ok=True)

    # Write timestamp to output file
    output_file_path = f"{output_folder}last_check_in.json"
    with open(output_file_path, "w") as f:
        json.dump(timestamp_data, f, indent=2)

    # Write _.syftperm file
    syftperm_data = {
        "admin": [client_config.email],
        "read": ["GLOBAL"],
        "write": [client_config.email],
        "filepath": f"{output_folder}_.syftperm",
        "terminal": False,
    }
    syftperm_path = f"{output_folder}_.syftperm"
    with open(syftperm_path, "w") as f:
        json.dump(syftperm_data, f, indent=2)

    print(f"Timestamp has been written to {output_file_path}")
    print(f"_.syftperm file has been written to {syftperm_path}")


if __name__ == "__main__":
    main()



## amke stats
# stat_folder = "./crypto/data"
# os.makedirs(stat_folder, exist_ok=True)
# stats_array = np.zeros(len(stats.keys())).astype(int)
# value = HE.encryptInt(stats_array)
# part_path = f"{stat_folder}/totals"
# value.save(part_path)

# # encode
# max_tv_id = 300_000 # random guess looking at their website
# tv_count_array = np.zeros(max_tv_id)

# tmdb_ids = np.array(list(value_counts.keys()))
# counts = np.array(value_counts.values)
# tv_count_array[tmdb_ids] += counts

# # decode
# non_zero_indices = np.nonzero(tv_count_array)[0].astype(int)
# non_zero_values = tv_count_array[non_zero_indices].astype(int)
# non_zero_dict = {int(k): int(v) for k, v in dict(zip(non_zero_indices, non_zero_values)).items()}
# print(non_zero_dict)
# return non_zero_dict
# HE = Pyfhel()
# HE.contextGen(scheme='bfv', n=2**15, t_bits=20)
# HE.keyGen()
# import os
# os.makedirs("crypto", exist_ok=True)
# HE.save_secret_key("crypto/pyfhel.secret")
# HE.save_public_key("crypto/pyfhel.pk")
# HE.save_context("crypto/pyfhel.context")

# brew install cmake zlib llvm libomp
# export CC=/opt/homebrew/opt/llvm/bin/clang
# uv pip install Pyfhel

# LDFLAGS="-L/opt/homebrew/opt/llvm/lib/c++ -L/opt/homebrew/opt/llvm/lib -lunwind"
# If you need to have llvm first in your PATH, run:
#   echo 'export PATH="/opt/homebrew/opt/llvm/bin:$PATH"' >> ~/.zshrc

# For compilers to find llvm you may need to set:
#   export LDFLAGS="-L/opt/homebrew/opt/llvm/lib"
#   export CPPFLAGS="-I/opt/homebrew/opt/llvm/include"

# ln -s /Users/madhavajay/dev/syft/users/madhava/me@madhavajay.com/public/datasets/netflix_tmdb_imdb/NetflixViewingHistory_TMDB_IMDB.csv ./dfs_0.csv
# /// script
# dependencies = [
#    "pandas==2.2.3",
#    "syftbox==0.1.0",
#    "phe==1.5.0",
#    "pyfhel==3.4.2",
# ]
#
# [tool.uv.sources]
# syftbox = { path = "/Users/madhavajay/dev/syft", editable = true }
# ///

__name__ = "netflix_stats"
__author__ = "madhava@openmined.org"


def input_reader(private: bool = False, datasite: str | None = None):
    import os

    import pandas as pd

    from syftbox.lib import extract_leftmost_email, sy_path

    df = None
    files = os.listdir("./inputs/dfs")
    for file in files:
        if file.startswith("dfs_"):
            full_path = f"./inputs/dfs/{file}"
            target_path = os.readlink(full_path)
            link_datasite = extract_leftmost_email(target_path)
            if link_datasite == datasite:
                df = pd.read_csv(sy_path(full_path, resolve_private=private))
                break
    inputs = {"datasite": datasite, "df": df}
    return inputs


def output_writer(result, private: bool = False):
    import json

    output_path = "./output/result/result.json"
    if not private:
        output_path = output_path.replace(".json", ".mock.json")
    with open(output_path, "w") as f:
        f.write(json.dumps(result))


# START YOUR CODE


def netflix_stats(datasite, df):
    import os

    crypto_folder = "./crypto"
    completed_sentinel = f"{crypto_folder}/{datasite}"
    if os.path.exists(completed_sentinel):
        print("✅ Already generated 🔐 Homomorphically Encrypted Stats")
        return

    import datetime
    import os

    import numpy as np
    import pandas as pd
    from Pyfhel import Pyfhel
    from Pyfhel.PyCtxt import PyCtxt

    HE = Pyfhel()
    HE.load_context(f"{crypto_folder}/pyfhel.context")
    HE.load_secret_key(f"{crypto_folder}/pyfhel.secret")
    HE.load_public_key(f"{crypto_folder}/pyfhel.pk")

    current_year = datetime.datetime.now().year
    df["netflix_date"] = pd.to_datetime(df["netflix_date"])
    year_df = df[df["netflix_date"].dt.year == current_year]
    year_tv_df = year_df[year_df["tmdb_media_type"] == "tv"]
    year_tv_df["day_of_week"] = year_tv_df["netflix_date"].dt.day_name()
    total_time = year_tv_df["imdb_runtime_minutes"].sum()
    total_views = len(year_tv_df)
    total_unique_show_views = year_tv_df["imdb_id"].nunique()
    # day_counts = year_tv_df["day_of_week"].value_counts()
    # favorite_day = list(day_counts.to_dict().keys())[0]
    # year_tv_df["day_of_week"] = year_tv_df["netflix_date"].dt.weekday
    # change to an int as a numpy array so we can add them

    value_counts = year_tv_df["tmdb_id"].value_counts().astype(int)
    # top_5_value_counts = {k: int(v) for k, v in value_counts.sort_values(ascending=False)[0:5].items()}

    stats = {
        "total_time": int(total_time),
        "total_views": int(total_views),
        "total_unique_show_views": int(total_unique_show_views),
        # "year_fav_day": str(favorite_day),
    }

    stat_folder = f"./{crypto_folder}/data"
    part_path = f"{stat_folder}/totals"
    slice_folder = f"{stat_folder}/view_counts"
    exists_files_folders = [stat_folder, part_path, slice_folder]
    os.makedirs(stat_folder, exist_ok=True)
    os.makedirs(slice_folder, exist_ok=True)

    for path in exists_files_folders:
        if not os.path.abspath(path):
            raise Exception(f"Requires {stat_folder} to finish syncing")

    imdb_id_files = os.listdir(slice_folder)
    if len(imdb_id_files) < 10:
        raise Exception(f"Requires {slice_folder} to finish syncing")

    # create totals
    stats_array = np.zeros(len(stats.keys())).astype(int)
    value = HE.encryptInt(stats_array)
    value.save(part_path)

    # write stats to encrypted array
    stats_array = np.zeros(len(stats)).astype(int)
    for i, value in enumerate(stats.values()):
        stats_array[i] = int(value)

    value = PyCtxt(pyfhel=HE)

    value.load(part_path)
    value += stats_array
    value.save(part_path)

    max_tv_id = 300_000  # just a guess
    slice_size = 30_000  # max size of the above HE context

    # create imdb_id slices
    counter = 0
    for i in range(0, max_tv_id + 1, slice_size):
        tv_count_array = np.zeros(slice_size).astype(int)
        tv_count_slice = HE.encryptInt(tv_count_array)
        part_path = f"{slice_folder}/tmdb_id_{counter:02}"
        tv_count_slice.save(part_path)
        counter += 1

    # write imdb_id value counts to chunked arrays
    for k, v in value_counts.items():
        imdb_id = int(k)
        index = imdb_id // slice_size
        sub_index = imdb_id % slice_size
        tv_count_slice = PyCtxt(pyfhel=HE)
        part_path = f"{slice_folder}/tmdb_id_{index:02}"
        empty_array = np.zeros(slice_size).astype(int)
        empty_array[sub_index] += int(v)
        tv_count_slice.load(part_path)
        tv_count_slice += empty_array
        tv_count_slice.save(part_path)

    with open(f"{crypto_folder}/{datasite}", "w") as f:
        print("✅ Writing 🔐 Homomorphically Encrypted Stats")
        f.write(str(datetime.datetime.now()))


# END YOUR CODE


def main():
    import argparse

    parser = argparse.ArgumentParser(description="Process some input.")
    parser.add_argument("--private", action="store_true", help="Run in private mode")
    parser.add_argument("--datasite", help="Datasite running operation")
    args = parser.parse_args()

    print(f"Running: {__name__} from {__author__}")
    inputs = input_reader(private=args.private, datasite=args.datasite)
    print("> Reading Inputs", inputs)

    output = netflix_stats(**inputs)

    print("> Writing Outputs", output)
    output_writer(output, private=args.private)
    print(f"> ✅ Running {__name__} Complete!")


main()



import os
from pathlib import Path

from syftbox.server.settings import ServerSettings


def test_server_settings_from_env():
    os.environ["SYFTBOX_DATA_FOLDER"] = "data_folder"
    os.environ["SYFTBOX_SNAPSHOT_FOLDER"] = "data_folder/snapshot_folder"
    os.environ["SYFTBOX_USER_FILE_PATH"] = "data_folder/user_file_path.json"

    settings = ServerSettings()
    print(settings)
    assert settings.data_folder == Path("data_folder")
    assert settings.snapshot_folder == Path("data_folder/snapshot_folder")
    assert settings.user_file_path == Path("data_folder/user_file_path.json")



import os
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from threading import Event

from loguru import logger
from watchdog.events import DirModifiedEvent

from syftbox.lib import (
    DirState,
    PermissionTree,
    bintostr,
    get_datasites,
    hash_dir,
    strtobin,
)
from syftbox.lib.lib import ClientConfig
from syftbox.server.models import FileChange, FileChangeKind

CLIENT_CHANGELOG_FOLDER = "syft_changelog"
CLIENT_APPS = "apps"
STAGING = "staging"
IGNORE_FOLDERS = [CLIENT_CHANGELOG_FOLDER, STAGING, CLIENT_APPS]


def get_ignore_rules(dir_state: DirState) -> list[str, str, str]:
    # get the ignore files
    syft_ignore_files = []
    folder_path = dir_state.sync_folder + "/" + dir_state.sub_path
    for afile, file_info in dir_state.tree.items():
        full_path = folder_path + "/" + afile
        sub_folder = os.path.dirname(full_path)

        if afile.endswith(".syftignore") and os.path.isfile(full_path):
            ignore_list = []
            with open(full_path) as f:
                ignore_list = f.readlines()
            for ignore_rule in ignore_list:
                ignore_rule = ignore_rule.strip()
                rule_prefix = sub_folder + "/" + ignore_rule
                syft_ignore_files.append((rule_prefix, sub_folder, afile))

    return syft_ignore_files


def filter_ignore_files(dir_state: DirState) -> DirState:
    # get the ignore files
    pruned_tree = dir_state.tree.copy()
    folder_path = dir_state.sync_folder + "/" + dir_state.sub_path
    syft_ignore_files = get_ignore_rules(dir_state)

    for rule_prefix, ignore_folder, ignore_file_path in syft_ignore_files:
        for afile, file_info in dir_state.tree.items():
            full_path = folder_path + "/" + afile
            if full_path.startswith(rule_prefix):
                # logger.info("> File ignored by .syftignore", afile, ignore_rule)
                if afile in pruned_tree:
                    del pruned_tree[afile]

    now = datetime.now().timestamp()
    return DirState(
        tree=pruned_tree,
        timestamp=now,
        sync_folder=dir_state.sync_folder,
        sub_path=dir_state.sub_path,
    )


# Recursive function to add folder structure
def add_to_folder_tree(leaf, parts):
    if not parts:
        return
    part = parts[0]
    if part not in leaf:
        leaf[part] = defaultdict(dict)
    add_to_folder_tree(leaf[part], parts[1:])


# Function to remove empty folders, working from deepest to shallowest
def remove_empty_folders(leaf, current_path, root_dir):
    # List all keys and attempt to remove empty subfolders first
    for folder in list(leaf.keys()):
        folder_path = os.path.join(current_path, folder)

        # If the folder contains subfolders, recursively check them
        if isinstance(leaf[folder], dict):
            remove_empty_folders(leaf[folder], folder_path, root_dir)

            # Now that we've processed the subfolders, check if it's empty on the filesystem
            full_path = root_dir + "/" + folder_path
            if os.path.isdir(full_path) and not os.listdir(full_path):
                os.rmdir(full_path)  # Remove the empty folder from the file system
                del leaf[folder]  # Remove it from the folder tree as well
            else:
                pass


# write operations
def diff_dirstate(old: DirState, new: DirState):
    sync_folder = old.sync_folder
    old_sub_path = old.sub_path
    try:
        changes = []
        for afile, file_info in new.tree.items():
            kind = None
            if afile in old.tree.keys():
                old_file_info = old.tree[afile]
                if (
                    old_file_info.file_hash != file_info.file_hash
                    and file_info.last_modified >= old_file_info.last_modified
                ):
                    # update
                    kind = FileChangeKind.WRITE
                else:
                    pass
                    # logger.info(
                    #     old_sub_path,
                    #     afile,
                    #     f"> 🔥 File hash eq=={old_file_info.file_hash == file_info.file_hash} "
                    #     f"or timestamp newer: {file_info.last_modified >= old_file_info.last_modified} "
                    #     f"dropping sync down {file_info}",
                    # )
            else:
                # create
                kind = FileChangeKind.CREATE

            if kind:
                change = FileChange(
                    kind=kind,
                    parent_path=old_sub_path,
                    sub_path=afile,
                    file_hash=file_info.file_hash,
                    last_modified=file_info.last_modified,
                    sync_folder=sync_folder,
                )
                changes.append(change)

        for afile, file_info in old.tree.items():
            if afile not in new.tree.keys():
                # delete
                now = datetime.now().timestamp()
                # TODO we need to overhaul this to prevent these kinds of edge cases
                SECS_SINCE_CHANGE = 5
                if now >= (file_info.last_modified + SECS_SINCE_CHANGE):
                    kind = FileChangeKind.DELETE
                    change = FileChange(
                        kind=kind,
                        parent_path=old.sub_path,
                        sub_path=afile,
                        file_hash=file_info.file_hash,
                        last_modified=file_info.last_modified,
                        sync_folder=sync_folder,
                    )
                    changes.append(change)
                else:
                    logger.info(
                        f"🔥 Skipping delete {afile} {file_info}. File change is < {SECS_SINCE_CHANGE} seconds ago"
                    )
        return changes
    except Exception as e:
        logger.info("Error in diff_dirstate", str(e))
        raise e


def prune_invalid_changes(new, valid_changes) -> DirState:
    new_tree = {}
    for file, file_info in new.tree.items():
        internal_path = new.sub_path + "/" + file
        if internal_path in valid_changes:
            new_tree[file] = file_info

    return DirState(
        tree=new_tree,
        timestamp=new.timestamp,
        sync_folder=new.sync_folder,
        sub_path=new.sub_path,
    )


def delete_files(new, deleted_files) -> DirState:
    new_tree = {}
    for file, file_info in new.tree.items():
        internal_path = new.sub_path + "/" + file
        if internal_path not in deleted_files:
            new_tree[file] = file_info

    return DirState(
        tree=new_tree,
        timestamp=new.timestamp,
        sync_folder=new.sync_folder,
        sub_path=new.sub_path,
    )


stop_event = Event()


stop_event = Event()

PLUGIN_NAME = "sync"


def filter_changes(
    user_email: str,
    changes: list[FileChange],
    perm_tree: PermissionTree,
):
    valid_changes = []
    valid_change_files = []
    invalid_changes = []
    invalid_permissions = []
    for change in changes:
        if perm_tree.has_corrupted_permission(change.full_path):
            invalid_permissions.append(change)
        elif change.kind in [
            FileChangeKind.WRITE,
            FileChangeKind.CREATE,
            FileChangeKind.DELETE,
        ]:
            perm_file_at_path = perm_tree.permission_for_path(change.full_path)
            if (
                user_email in perm_file_at_path.write
                or "GLOBAL" in perm_file_at_path.write
            ) or user_email in perm_file_at_path.admin:
                valid_changes.append(change)
                valid_change_files.append(change.sub_path)
                continue
            # # todo we need to handle this properly
            # if perm_file_at_path.admin == [user_email]:
            #     if change.internal_path.endswith("_.syftperm"):
            #         # include changes for syft_perm file even if only we have read perms.
            #         valid_changes.append(change)
            #         valid_change_files.append(change.sub_path)
            #         continue

        else:
            invalid_changes.append(change)
    return valid_changes, valid_change_files, invalid_changes, invalid_permissions


def push_changes(
    client_config: ClientConfig, changes: list[FileChange]
) -> list[FileChange]:
    written_changes = []
    for change in changes:
        try:
            data = {
                "email": client_config.email,
                "change": change.model_dump(mode="json"),
            }
            if change.kind_write:
                if os.path.isdir(change.full_path):
                    # Handle directory
                    data["is_directory"] = True
                else:
                    # Handle file
                    data["data"] = bintostr(change.read())
            elif change.kind_delete:
                # no data for delete operations
                pass

            response = client_config.server_client.post(
                "/write",
                json=data,
            )
            write_response = response.json()
            change_result = write_response["change"]
            change_result["kind"] = FileChangeKind(change_result["kind"])
            ok_change = FileChange(**change_result)
            if response.status_code == 200:
                if "accepted" in write_response and write_response["accepted"]:
                    written_changes.append(ok_change)
                else:
                    logger.info(f"> 🔥 Rejected change: {change.full_path}", ok_change)
            else:
                logger.info(
                    f"> {client_config.email} FAILED /write {change.kind} {change.internal_path}",
                )
        except Exception as e:
            logger.info(
                f"Failed to call /write on the server for {change.internal_path}",
                str(e),
            )
    return written_changes


def pull_changes(client_config, changes: list[FileChange]):
    remote_changes = []
    for change in changes:
        try:
            data = {
                "email": client_config.email,
                "change": change.model_dump(mode="json"),
            }
            response = client_config.server_client.post(
                "/read",
                json=data,
            )
            read_response = response.json()
            change_result = read_response["change"]
            change_result["kind"] = FileChangeKind(change_result["kind"])
            ok_change = FileChange(**change_result)

            if ok_change.kind_write:
                if read_response.get("is_directory", False):
                    data = None
                else:
                    data = strtobin(read_response["data"])
            elif change.kind_delete:
                data = None

            if response.status_code == 200:
                remote_changes.append((ok_change, data))
            else:
                logger.info(
                    f"> {client_config.email} FAILED /read {change.kind} {change.internal_path}",
                )
        except Exception as e:
            logger.error("Failed to call /read on the server")
            logger.exception(e)
    return remote_changes


def list_datasites(client_config: ClientConfig):
    datasites = []
    try:
        response = client_config.server_client.get(
            "/list_datasites",
        )
        read_response = response.json()
        remote_datasites = read_response["datasites"]

        if response.status_code == 200:
            datasites = remote_datasites
        else:
            logger.info(f"> {client_config.email} FAILED /list_datasites")
    except Exception as e:
        logger.error("Failed to call /list_datasites on the server")
        logger.exception(e)
    return datasites


def get_remote_state(client_config: ClientConfig, sub_path: str):
    try:
        data = {
            "email": client_config.email,
            "sub_path": sub_path,
        }

        response = client_config.server_client.post(
            "/dir_state",
            json=data,
        )
        try:
            state_response = response.json()
        except Exception:
            logger.error(f"""Failed to call /dir_state for {sub_path} response Not JSON: {response.text}. \
This may be related to broken (empty!) syftperm files""")
            return None

        if response.status_code == 200:
            if isinstance(state_response, dict) and "dir_state" in state_response:
                dir_state = DirState(**state_response["dir_state"])
                fix_tree = {}
                for key, value in dir_state.tree.items():
                    fix_tree[key] = value
                dir_state.tree = fix_tree
                return dir_state
            else:
                logger.info(
                    "/dir_state returned a bad result",
                    type(state_response),
                    state_response,
                )
        logger.info(f"> {client_config.email} FAILED /dir_state: {sub_path}")
        return None
    except Exception as e:
        logger.error("Failed to call /dir_state on the server")
        logger.exception(e)


def create_datasites(client_config):
    datasites = list_datasites(client_config)
    for datasite in datasites:
        # get the top level perm file
        os.makedirs(os.path.join(client_config.sync_folder, datasite), exist_ok=True)


def ascii_for_change(changes) -> str:
    count = 0
    change_text = ""
    for change in changes:
        count += 1
        pipe = "├──"
        if count == len(changes):
            pipe = "└──"
        change_text += pipe + change + "\n"
    return change_text


def handle_empty_folders(client_config, datasite):
    changes = []
    datasite_path = os.path.join(client_config.sync_folder, datasite)

    for root, dirs, files in os.walk(datasite_path):
        if not files and not dirs:
            # This is an empty folder
            relative_path = os.path.relpath(root, datasite_path)
            if relative_path == ".":
                continue  # Skip the root folder

            change = FileChange(
                kind=FileChangeKind.CREATE,
                parent_path=datasite,
                sub_path=relative_path,
                file_hash="",  # Empty folders don't have a hash
                last_modified=os.path.getmtime(root),
                sync_folder=client_config.sync_folder,
            )
            changes.append(change)

    return changes


def filter_changes_ignore(
    pre_filter_changes: list[FileChange], syft_ignore_files
) -> list[FileChange]:
    filtered_changes = []
    for change in pre_filter_changes:
        keep = True
        for syft_ignore in syft_ignore_files:
            if change.full_path.startswith(syft_ignore[0]):
                keep = False
                break
        if keep:
            filtered_changes.append(change)

    return filtered_changes


def sync_up(client_config: ClientConfig):
    # create a folder to store the change log
    change_log_folder = f"{client_config.sync_folder}/{CLIENT_CHANGELOG_FOLDER}"
    os.makedirs(change_log_folder, exist_ok=True)

    # get all the datasites
    datasites = get_datasites(client_config.sync_folder)

    n_changes = 0

    for datasite in datasites:
        # get the top level perm file
        datasite_path = os.path.join(client_config.sync_folder, datasite)

        perm_tree = PermissionTree.from_path(datasite_path)

        dir_filename = f"{change_log_folder}/{datasite}.json"

        # get the old dir state
        old_dir_state = None
        try:
            # it might not exist yet
            old_dir_state = DirState.load(dir_filename)
            fix_tree = {}
            for key, value in old_dir_state.tree.items():
                fix_tree[key] = value
            old_dir_state.tree = fix_tree
        except Exception:
            pass

        if old_dir_state is None:
            old_dir_state = DirState(
                tree={},
                timestamp=0,
                sync_folder=client_config.sync_folder,
                sub_path=datasite,
            )

        # get the new dir state
        unfiltered_new_dir_state = hash_dir(
            client_config.sync_folder, datasite, IGNORE_FOLDERS
        )

        # ignore files
        syft_ignore_files = get_ignore_rules(unfiltered_new_dir_state)
        new_dir_state = filter_ignore_files(unfiltered_new_dir_state)

        pre_filter_changes = diff_dirstate(old_dir_state, new_dir_state)

        # Add handling for empty folders
        empty_folder_changes = handle_empty_folders(client_config, datasite)
        pre_filter_changes.extend(empty_folder_changes)

        changes = filter_changes_ignore(pre_filter_changes, syft_ignore_files)

        if len(changes) == 0:
            continue

        val, val_files, inval_changes, inval_permissions = filter_changes(
            client_config.email, changes, perm_tree
        )
        if len(inval_permissions) > 0:
            logger.warning(
                f"Filtered {len(inval_permissions)} changes with corrupted permissions"
            )
            inval_permission_files = [
                change.internal_path for change in inval_permissions
            ]
            logger.debug(
                f"Filtered changes with corrupted permissions: {inval_permission_files}"
            )

        # send val changes
        results = push_changes(client_config, val)

        deleted_files = []
        changed_files = []
        for result in results:
            if result.kind_write:
                changed_files.append(result.internal_path)
            elif result.kind_delete:
                deleted_files.append(result.internal_path)

        synced_dir_state = prune_invalid_changes(new_dir_state, changed_files)

        # combine successful changes qwith old dir state
        combined_tree = old_dir_state.tree

        # add new successful changes
        combined_tree.update(synced_dir_state.tree)
        synced_dir_state.tree = combined_tree

        synced_dir_state = delete_files(synced_dir_state, deleted_files)

        change_text = ""
        if len(changed_files):
            change_text += f"🔼 Syncing Up {len(changed_files)} Changes\n"
            change_text += ascii_for_change(changed_files)

        if len(deleted_files):
            change_text += f"❌ Syncing Up {len(deleted_files)} Deletes\n"
            change_text += ascii_for_change(deleted_files)

        synced_dir_state.save(dir_filename)
        n_changes += len(changed_files) + len(deleted_files)

    return n_changes


def sync_down(client_config) -> int:
    # create a folder to store the change log
    change_log_folder = f"{client_config.sync_folder}/{CLIENT_CHANGELOG_FOLDER}"
    os.makedirs(change_log_folder, exist_ok=True)

    # get all the datasites
    datasites = get_datasites(client_config.sync_folder)

    with ThreadPoolExecutor(max_workers=min(6, len(datasites))) as executor:
        results = []
        for datasite in datasites:
            n_changes_datasite_future = executor.submit(
                sync_down_dataite, datasite, client_config, change_log_folder
            )
            results.append(n_changes_datasite_future)
        n_changes = sum([x.result() for x in results])

    return n_changes


def sync_down_dataite(datasite, client_config, change_log_folder):
    # get the top level perm file

    dir_filename = f"{change_log_folder}/{datasite}.json"

    # datasite_path = os.path.join(client_config.sync_folder, datasite)

    # perm_tree = PermissionTree.from_path(datasite_path)

    # get the new dir state

    unfiltered_new_dir_state = hash_dir(
        client_config.sync_folder, datasite, IGNORE_FOLDERS
    )
    syft_ignore_files = get_ignore_rules(unfiltered_new_dir_state)

    # ignore files
    new_dir_state = filter_ignore_files(unfiltered_new_dir_state)

    remote_dir_state = get_remote_state(client_config, datasite)
    if not remote_dir_state:
        logger.info(
            f"Could not find remote state for {datasite}, skipping syncing down"
        )
        return 0

    pre_filter_changes = diff_dirstate(new_dir_state, remote_dir_state)

    # Add handling for empty folders
    empty_folder_changes = handle_empty_folders(client_config, datasite)
    pre_filter_changes.extend(empty_folder_changes)

    changes = filter_changes_ignore(pre_filter_changes, syft_ignore_files)

    if len(changes) == 0:
        return 0

    # fetch writes from the /read endpoint
    fetch_files = []
    for change in changes:
        if change.kind_write:
            fetch_files.append(change)

    results = pull_changes(client_config, fetch_files)

    # make writes
    changed_files = []
    for change, data in results:
        change.sync_folder = client_config.sync_folder
        if change.kind_write:
            if data is None:  # This is an empty directory
                os.makedirs(change.full_path, exist_ok=True)
                changed_files.append(change.internal_path)
            else:
                result = change.write(data)
                if result:
                    changed_files.append(change.internal_path)

    # delete local files dont need the server
    deleted_files = []
    for change in changes:
        change.sync_folder = client_config.sync_folder
        if change.kind_delete:
            # perm_file_at_path = perm_tree.permission_for_path(change.sub_path)
            # if client_config.email in perm_file_at_path.admin:
            #     continue
            result = change.delete()
            if result:
                deleted_files.append(change.internal_path)

    # remove empty folders
    folder_tree = defaultdict(dict)
    # Process each file and build the folder structure
    for item in deleted_files:
        folders = os.path.dirname(item).split("/")
        add_to_folder_tree(folder_tree, folders)

    # Remove empty folders, starting from the root directory
    remove_empty_folders(folder_tree, "/", root_dir=client_config.sync_folder)

    synced_dir_state = prune_invalid_changes(new_dir_state, changed_files)

    # combine successfulc hanges qwith old dir state
    # we use unfiltered so they keep being ignored but we could change these to another list?
    combined_tree = unfiltered_new_dir_state.tree
    combined_tree.update(synced_dir_state.tree)
    synced_dir_state.tree = combined_tree

    synced_dir_state = delete_files(synced_dir_state, deleted_files)

    change_text = ""
    if len(changed_files):
        change_text += f"⏬ Syncing Down {len(changed_files)} Changes\n"
        change_text += ascii_for_change(changed_files)
    if len(deleted_files):
        change_text += f"❌ Syncing Down {len(deleted_files)} Deletes\n"
        change_text += ascii_for_change(deleted_files)

    if len(change_text) > 0:
        import threading

        logger.info(f"{threading.get_ident()} {os.getpid()} {change_text}")

    synced_dir_state.save(dir_filename)
    return len(changed_files) + len(deleted_files)


SYNC_UP_ENABLED = True
SYNC_DOWN_ENABLED = True


def do_sync(shared_state):
    event_length = len(shared_state.fs_events)
    shared_state.fs_events = []
    try:
        if not stop_event.is_set():
            num_changes = 0
            if shared_state.client_config.token:
                try:
                    create_datasites(shared_state.client_config)
                except Exception as e:
                    logger.error("failed to get_datasites", e)
                    logger.exception(e)

                try:
                    if SYNC_UP_ENABLED:
                        num_changes += sync_up(shared_state.client_config)
                    else:
                        logger.info("❌ Sync Up Disabled")
                except Exception as e:
                    logger.error("failed to sync up", e)
                    logger.exception(e)

                try:
                    if SYNC_DOWN_ENABLED:
                        num_changes += sync_down(shared_state.client_config)
                    else:
                        logger.info("❌ Sync Down Disabled")
                except Exception as e:
                    logger.error("failed to sync down", e)
                    logger.exception(e)
            if num_changes == 0:
                if event_length:
                    logger.info(f"✅ Synced {event_length} File Events")
                else:
                    logger.info("✅ Synced due to Timer")
    except Exception as e:
        logger.error("Failed to run plugin")
        logger.exception(e)


FLUSH_SYNC_TIMEOUT = 0.5
DEFAULT_SCHEDULE = 1000


def run(shared_state, *args, **kwargs):
    if len(args) == 1:
        event = args[0]
        # ignore certain files / folders
        if hasattr(event, "src_path"):
            if CLIENT_CHANGELOG_FOLDER in event.src_path:
                return

        # ignore these events for now on linux
        # FileOpenedEvent
        # FileClosedNoWriteEvent
        # DirModifiedEvent
        if event.event_type in ["opened", "closed_no_write"]:
            return

        if isinstance(event, DirModifiedEvent):
            return

        shared_state.fs_events.append(event)

    do_sync(shared_state)



import argparse
import atexit
import contextlib
import importlib
import os
import platform
import sys
import time
import types
from dataclasses import dataclass
from datetime import datetime
from functools import partial
from pathlib import Path

import uvicorn
from apscheduler.jobstores.sqlalchemy import SQLAlchemyJobStore
from apscheduler.schedulers.background import BackgroundScheduler
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from fastapi import Body, FastAPI, HTTPException, Request
from fastapi.encoders import jsonable_encoder
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import FileResponse, HTMLResponse, JSONResponse
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates
from loguru import logger
from pydantic import BaseModel
from typing_extensions import Any

from syftbox import __version__
from syftbox.client.utils.error_reporting import make_error_report
from syftbox.lib import (
    DEFAULT_CONFIG_PATH,
    ClientConfig,
    SharedState,
    load_or_create_config,
)
from syftbox.lib.logger import zip_logs


class CustomFastAPI(FastAPI):
    loaded_plugins: dict
    running_plugins: dict
    scheduler: Any
    shared_state: dict
    job_file: str
    watchdog: Any
    job_file: str


current_dir = Path(__file__).parent
# Initialize FastAPI app and scheduler

templates = Jinja2Templates(directory=str(current_dir / "templates"))


PLUGINS_DIR = current_dir / "plugins"
sys.path.insert(0, os.path.dirname(PLUGINS_DIR))

DEFAULT_SYNC_FOLDER = os.path.expanduser("~/Desktop/SyftBox")


ASSETS_FOLDER = current_dir.parent / "assets"
ICON_FOLDER = ASSETS_FOLDER / "icon"

WATCHDOG_IGNORE = ["apps"]


@dataclass
class Plugin:
    name: str
    module: types.ModuleType
    schedule: int
    description: str


def process_folder_input(user_input, default_path):
    if not user_input:
        return default_path
    if "/" not in user_input:
        # User only provided a folder name, use it with the default parent path
        parent_path = os.path.dirname(default_path)
        return os.path.join(parent_path, user_input)
    return os.path.expanduser(user_input)


def initialize_shared_state(client_config: ClientConfig) -> SharedState:
    shared_state = SharedState(client_config=client_config)
    return shared_state


def load_plugins(client_config: ClientConfig) -> dict[str, Plugin]:
    loaded_plugins = {}
    if os.path.exists(PLUGINS_DIR) and os.path.isdir(PLUGINS_DIR):
        for item in os.listdir(PLUGINS_DIR):
            if item.endswith(".py") and not item.startswith("__"):
                plugin_name = item[:-3]
                try:
                    module = importlib.import_module(f"plugins.{plugin_name}")
                    schedule = getattr(
                        module,
                        "DEFAULT_SCHEDULE",
                        5000,
                    )  # Default to 5000ms if not specified
                    description = getattr(
                        module,
                        "DESCRIPTION",
                        "No description available.",
                    )
                    plugin = Plugin(
                        name=plugin_name,
                        module=module,
                        schedule=schedule,
                        description=description,
                    )
                    loaded_plugins[plugin_name] = plugin
                except Exception as e:
                    logger.info(e)

    return loaded_plugins


def generate_key_pair() -> tuple[bytes, bytes]:
    private_key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=2048,
        backend=default_backend(),
    )
    public_key = private_key.public_key()

    private_pem = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption(),
    )
    public_pem = public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo,
    )

    return private_pem, public_pem


def is_valid_datasite_name(name):
    return name.isalnum() or all(c.isalnum() or c in ("-", "_") for c in name)


# API Models
class PluginRequest(BaseModel):
    plugin_name: str


class SharedStateRequest(BaseModel):
    key: str
    value: str


class DatasiteRequest(BaseModel):
    name: str


# Function to be scheduled
def run_plugin(plugin_name, *args, **kwargs):
    try:
        module = app.loaded_plugins[plugin_name].module
        module.run(app.shared_state, *args, **kwargs)
    except Exception as e:
        logger.exception(e)


def start_plugin(app: CustomFastAPI, plugin_name: str):
    if plugin_name not in app.loaded_plugins:
        raise HTTPException(
            status_code=400,
            detail=f"Plugin {plugin_name} is not loaded",
        )

    if plugin_name in app.running_plugins:
        raise HTTPException(
            status_code=400,
            detail=f"Plugin {plugin_name} is already running",
        )

    try:
        plugin = app.loaded_plugins[plugin_name]

        existing_job = app.scheduler.get_job(plugin_name)
        if existing_job is None:
            job = app.scheduler.add_job(
                func=run_plugin,
                trigger="interval",
                seconds=plugin.schedule / 1000,
                id=plugin_name,
                args=[plugin_name],
            )
            app.running_plugins[plugin_name] = {
                "job": job,
                "start_time": time.time(),
                "schedule": plugin.schedule,
            }
            return {"message": f"Plugin {plugin_name} started successfully"}
        else:
            logger.info(f"Job {existing_job}, already added")
            return {"message": f"Plugin {plugin_name} already started"}
    except Exception as e:
        raise HTTPException(
            status_code=500,
            detail=f"Failed to start plugin {plugin_name}: {e!s}",
        )


# Parsing arguments and initializing shared state
def parse_args():
    parser = argparse.ArgumentParser(
        description="Run the web application with plugins.",
    )
    parser.add_argument(
        "--config_path", type=str, default=DEFAULT_CONFIG_PATH, help="config path"
    )

    parser.add_argument("--debug", action="store_true", help="debug mode")

    parser.add_argument("--sync_folder", type=str, help="sync folder path")
    parser.add_argument("--email", type=str, help="email")
    parser.add_argument("--port", type=int, default=8080, help="Port number")
    parser.add_argument(
        "--server",
        type=str,
        default="https://syftbox.openmined.org",
        help="Server",
    )
    subparsers = parser.add_subparsers(dest="command", help="Sub-command help")
    start_parser = subparsers.add_parser("report", help="Generate an error report")
    start_parser.add_argument(
        "--path",
        type=str,
        help="Path to the error report file",
        default=f"./syftbox_logs_{datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}",
    )

    return parser.parse_args()


# def start_watchdog(app) -> FSWatchdog:
#     def sync_on_event(event: FileSystemEvent):
#         run_plugin("sync", event)

#     watch_dir = Path(app.shared_state.client_config.sync_folder)
#     watch_dir.mkdir(parents=True, exist_ok=True)
#     event_handler = AnyFileSystemEventHandler(
#         watch_dir,
#         callbacks=[sync_on_event],
#         ignored=WATCHDOG_IGNORE,
#     )
#     watchdog = FSWatchdog(watch_dir, event_handler)
#     watchdog.start()
#     return watchdog


@contextlib.asynccontextmanager
async def lifespan(app: CustomFastAPI, client_config: ClientConfig | None = None):
    # Startup
    logger.info(
        f"> Starting SyftBox Client: {__version__} Python {platform.python_version()}"
    )

    config_path = os.environ.get("SYFTBOX_CLIENT_CONFIG_PATH")
    if config_path:
        client_config = ClientConfig.load(config_path)

    # client_config needs to be closed if it was created in this context
    # if it is passed as lifespan arg (eg for testing) it should be managed by the caller instead.
    close_client_config: bool = False
    if client_config is None:
        args = parse_args()
        client_config = load_or_create_config(args)
        close_client_config = True
    app.shared_state = SharedState(client_config=client_config)

    # Clear the lock file on the first run if it exists
    job_file = client_config.config_path.replace(".json", ".sql")
    app.job_file = job_file
    if os.path.exists(job_file):
        os.remove(job_file)
        logger.info(f"> Cleared existing job file: {job_file}")

    # Start the scheduler
    jobstores = {"default": SQLAlchemyJobStore(url=f"sqlite:///{job_file}")}
    scheduler = BackgroundScheduler(jobstores=jobstores)
    scheduler.start()
    atexit.register(partial(stop_scheduler, app))

    app.scheduler = scheduler
    app.running_plugins = {}
    app.loaded_plugins = load_plugins(client_config)
    logger.info("> Loaded plugins:", sorted(list(app.loaded_plugins.keys())))
    # app.watchdog = start_watchdog(app)

    logger.info("> Starting autorun plugins:", sorted(client_config.autorun_plugins))
    for plugin in client_config.autorun_plugins:
        start_plugin(app, plugin)

    yield  # This yields control to run the application

    logger.info("> Shutting down...")
    scheduler.shutdown()
    # app.watchdog.stop()
    if close_client_config:
        client_config.close()


def stop_scheduler(app: FastAPI):
    # Remove the lock file if it exists
    if os.path.exists(app.job_file):
        os.remove(app.job_file)
        logger.info("> Scheduler stopped and lock file removed.")


app: CustomFastAPI = FastAPI(lifespan=lifespan)

app.mount("/static", StaticFiles(directory=current_dir / "static"), name="static")


# Add CORS middleware
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],  # Allows all origins
    allow_credentials=True,
    allow_methods=["*"],  # Allows all methods
    allow_headers=["*"],  # Allows all headers
)


@app.get("/", response_class=HTMLResponse)
async def plugin_manager(request: Request):
    # Pass the request to the template to allow FastAPI to render it
    return templates.TemplateResponse("index.html", {"request": request})


@app.get("/client_email")
def get_client_email():
    try:
        email = app.shared_state.client_config.email
        return JSONResponse(content={"email": email})
    except AttributeError as e:
        raise HTTPException(
            status_code=500,
            detail=f"Error accessing client email: {e!s}",
        )


@app.get("/state")
def get_shared_state():
    return JSONResponse(content=app.shared_state.data)


@app.get("/datasites")
def list_datasites():
    datasites = app.shared_state.get("my_datasites", [])
    # Use jsonable_encoder to encode the datasites object
    return JSONResponse(content={"datasites": jsonable_encoder(datasites)})


# FastAPI Routes
@app.get("/plugins")
def list_plugins():
    plugins = [
        {
            "name": plugin_name,
            "default_schedule": plugin.schedule,
            "is_running": plugin_name in app.running_plugins,
            "description": plugin.description,
        }
        for plugin_name, plugin in app.loaded_plugins.items()
    ]
    return {"plugins": plugins}


@app.post("/launch")
def launch_plugin(plugin_request: PluginRequest, request: Request):
    return start_plugin(request.app, plugin_request.plugin_name)


@app.get("/running")
def list_running_plugins():
    running = {
        name: {
            "is_running": data["job"].next_run_time is not None,
            "run_time": time.time() - data["start_time"],
            "schedule": data["schedule"],
        }
        for name, data in app.running_plugins.items()
    }
    return {"running_plugins": running}


@app.post("/kill")
def kill_plugin(request: PluginRequest):
    plugin_name = request.plugin_name

    if plugin_name not in app.running_plugins:
        raise HTTPException(
            status_code=400,
            detail=f"Plugin {plugin_name} is not running",
        )

    try:
        app.scheduler.remove_job(plugin_name)
        plugin_module = app.loaded_plugins[plugin_name].module
        if hasattr(plugin_module, "stop"):
            plugin_module.stop()
        del app.running_plugins[plugin_name]
        return {"message": f"Plugin {plugin_name} stopped successfully"}
    except Exception as e:
        raise HTTPException(
            status_code=500,
            detail=f"Failed to stop plugin {plugin_name}: {e!s}",
        )


@app.post("/file_operation")
async def file_operation(
    operation: str = Body(...),
    file_path: str = Body(...),
    content: str = Body(None),
):
    full_path = Path(app.shared_state.client_config.sync_folder) / file_path

    # Ensure the path is within the SyftBox directory
    if not full_path.resolve().is_relative_to(
        Path(app.shared_state.client_config.sync_folder),
    ):
        raise HTTPException(
            status_code=403,
            detail="Access to files outside SyftBox directory is not allowed",
        )

    if operation == "read":
        if not full_path.is_file():
            raise HTTPException(status_code=404, detail="File not found")
        return FileResponse(full_path)

    if operation in ["write", "append"]:
        if content is None:
            raise HTTPException(
                status_code=400,
                detail="Content is required for write or append operation",
            )

        # Ensure the directory exists
        full_path.parent.mkdir(parents=True, exist_ok=True)

        try:
            mode = "w" if operation == "write" else "a"
            with open(full_path, mode) as f:
                f.write(content)
            return JSONResponse(content={"message": f"File {operation}ed successfully"})
        except Exception as e:
            raise HTTPException(
                status_code=500,
                detail=f"Failed to {operation} file: {e!s}",
            )

    else:
        raise HTTPException(
            status_code=400,
            detail="Invalid operation. Use 'read', 'write', or 'append'",
        )


def get_syftbox_src_path():
    import importlib.util

    module_name = "syftbox"
    spec = importlib.util.find_spec(module_name)
    return spec.origin


def main() -> None:
    args = parse_args()
    client_config = load_or_create_config(args)
    error_config = make_error_report(client_config)

    if args.command == "report":
        output_path = Path(args.path).resolve()
        output_path_with_extension = zip_logs(output_path)
        logger.info(f"Logs saved to: {output_path_with_extension}.")
        logger.info("Please share your bug report together with the zipped logs")
        return

    logger.info(f"Client metadata: {error_config.model_dump_json(indent=2)}")

    os.environ["SYFTBOX_DATASITE"] = client_config.email
    os.environ["SYFTBOX_CLIENT_CONFIG_PATH"] = client_config.config_path

    logger.info("Dev Mode: ", os.environ.get("SYFTBOX_DEV"))
    logger.info("Wheel: ", os.environ.get("SYFTBOX_WHEEL"))

    debug = True if args.debug else False
    port = client_config.port
    max_attempts = 10  # Maximum number of port attempts

    for attempt in range(max_attempts):
        try:
            uvicorn.run(
                "syftbox.client.client:app" if debug else app,
                host="0.0.0.0",
                port=port,
                log_level="debug" if debug else "info",
                reload=debug,
                reload_dirs="./syftbox",
            )
            return  # If successful, exit the loop
        except SystemExit as e:
            if e.code != 1:  # If it's not the "Address already in use" error
                raise
            logger.info(f"Failed to start server on port {port}. Trying next port.")
            port = 0
    logger.info(f"Unable to find an available port after {max_attempts} attempts.")
    sys.exit(1)


if __name__ == "__main__":
    main()



import subprocess
from pathlib import Path

ASSETS_FOLDER = Path(__file__).parents[2] / "assets"
ICONS_PKG = ASSETS_FOLDER / "icon.zip"


# Function to search for Icon\r file
def search_icon_file(src_path: Path) -> Path | None:
    if not src_path.exists():
        return None
    for file_path in src_path.iterdir():
        if "Icon" in file_path.name and "\r" in file_path.name:
            return file_path


# if you knew the pain of this function
def find_icon_file(src_path: Path) -> Path:
    # First attempt to find the Icon\r file
    icon_file = search_icon_file(src_path)
    if icon_file:
        return icon_file

    if not ICONS_PKG.exists():
        # If still not found, raise an error
        raise FileNotFoundError(f"{ICONS_PKG} not found")

    try:
        # cant use other zip tools as they don't unpack it correctly
        subprocess.run(
            ["ditto", "-xk", str(ICONS_PKG), str(src_path.parent)],
            check=True,
        )

        # Try to find the Icon\r file again after extraction
        icon_file = search_icon_file(src_path)
        if icon_file:
            return icon_file
    except subprocess.CalledProcessError:
        raise RuntimeError("Failed to unzip icon.zip using macOS CLI tool.")


def copy_icon_file(icon_folder: str, dest_folder: str) -> None:
    dest_path = Path(dest_folder)
    icon_path = Path(icon_folder)
    src_icon_path = find_icon_file(icon_path)
    if not dest_path.exists():
        raise FileNotFoundError(f"Destination folder '{dest_folder}' does not exist.")

    # shutil wont work with these special icon files
    subprocess.run(["cp", "-p", src_icon_path, dest_folder], check=True)
    subprocess.run(["SetFile", "-a", "C", dest_folder], check=True)



import argparse
import contextlib
import json
import os
import platform
import random
import sys
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path

import uvicorn
from fastapi import Depends, FastAPI, HTTPException, Request
from fastapi.responses import (
    FileResponse,
    HTMLResponse,
    JSONResponse,
    PlainTextResponse,
    RedirectResponse,
)
from jinja2 import Template
from loguru import logger
from typing_extensions import Any, Optional

from syftbox.__version__ import __version__
from syftbox.lib import (
    Jsonable,
    PermissionTree,
    bintostr,
    filter_read_state,
    get_datasites,
    hash_dir,
    strtobin,
)
from syftbox.server.models import (
    DirStateRequest,
    DirStateResponse,
    ListDatasitesResponse,
    ReadRequest,
    ReadResponse,
    WriteRequest,
    WriteResponse,
)
from syftbox.server.settings import ServerSettings, get_server_settings

current_dir = Path(__file__).parent


def load_list(cls, filepath: str) -> list[Any]:
    try:
        with open(filepath) as f:
            data = f.read()
            d = json.loads(data)
            ds = []
            for di in d:
                ds.append(cls(**di))
            return ds
    except Exception as e:
        logger.info(f"Unable to load list file: {filepath}. {e}")
    return None


def save_list(obj: Any, filepath: str) -> None:
    dicts = []
    for d in obj:
        dicts.append(d.to_dict())
    with open(filepath, "w") as f:
        f.write(json.dumps(dicts))


def load_dict(cls, filepath: str) -> list[Any]:
    try:
        with open(filepath) as f:
            data = f.read()
            d = json.loads(data)
            dicts = {}
            for key, value in d.items():
                dicts[key] = cls(**value)
            return dicts
    except Exception as e:
        logger.info(f"Unable to load dict file: {filepath}. {e}")
    return None


def save_dict(obj: Any, filepath: str) -> None:
    dicts = {}
    for key, value in obj.items():
        dicts[key] = value.to_dict()

    with open(filepath, "w") as f:
        f.write(json.dumps(dicts))


@dataclass
class User(Jsonable):
    email: str
    token: int  # TODO


class Users:
    def __init__(self, path: Path) -> None:
        self.path = path
        self.users = {}
        self.load()

    def load(self):
        if os.path.exists(str(self.path)):
            users = load_dict(User, str(self.path))
        else:
            users = None
        if users:
            self.users = users

    def save(self):
        save_dict(self.users, str(self.path))

    def get_user(self, email: str) -> Optional[User]:
        if email not in self.users:
            return None
        return self.users[email]

    def create_user(self, email: str) -> int:
        if email in self.users:
            # for now just return the token
            return self.users[email].token
            # raise Exception(f"User already registered: {email}")
        token = random.randint(0, sys.maxsize)
        user = User(email=email, token=token)
        self.users[email] = user
        self.save()
        return token

    def __repr__(self) -> str:
        string = ""
        for email, user in self.users.items():
            string += f"{email}: {user}"
        return string


def get_users(request: Request) -> Users:
    return request.state.users


def create_folders(folders: list[str]) -> None:
    for folder in folders:
        if not os.path.exists(folder):
            os.makedirs(folder, exist_ok=True)


@contextlib.asynccontextmanager
async def lifespan(app: FastAPI, settings: ServerSettings | None = None):
    # Startup
    logger.info(
        f"> Starting SyftBox Server {__version__}. Python {platform.python_version()}"
    )
    if settings is None:
        settings = ServerSettings()
    logger.info(settings)

    logger.info("> Creating Folders")

    create_folders(settings.folders)

    users = Users(path=settings.user_file_path)
    logger.info("> Loading Users")
    logger.info(users)

    yield {
        "server_settings": settings,
        "users": users,
    }

    logger.info("> Shutting down server")


app = FastAPI(lifespan=lifespan)
# Define the ASCII art
ascii_art = rf"""
 ____         __ _   ____
/ ___| _   _ / _| |_| __ )  _____  __
\___ \| | | | |_| __|  _ \ / _ \ \/ /
 ___) | |_| |  _| |_| |_) | (_) >  <
|____/ \__, |_|  \__|____/ \___/_/\_\
       |___/        {__version__:>17}


# Install Syftbox (MacOS and Linux)
curl -LsSf https://syftbox.openmined.org/install.sh | sh

# Run the client
syftbox client
"""


@app.get("/", response_class=PlainTextResponse)
async def get_ascii_art(request: Request):
    req_host = request.headers.get("host", "")
    if "syftboxstage" in req_host:
        return ascii_art.replace("syftbox.openmined.org", "syftboxstage.openmined.org")
    return ascii_art


@app.get("/wheel/{path:path}", response_class=HTMLResponse)
async def get_wheel(path: str):
    if path == "":  # Check if path is empty (meaning "/datasites/")
        return RedirectResponse(url="/")

    filename = path.split("/")[0]
    if filename.endswith(".whl"):
        wheel_path = os.path.expanduser("~/syftbox-0.1.0-py3-none-any.whl")
        return FileResponse(wheel_path, media_type="application/octet-stream")
    return filename


def get_file_list(directory: str | Path = ".") -> list[dict[str, Any]]:
    # TODO rewrite with pathlib
    directory = str(directory)

    file_list = []
    for item in os.listdir(directory):
        item_path = os.path.join(directory, item)
        is_dir = os.path.isdir(item_path)
        size = os.path.getsize(item_path) if not is_dir else "-"
        mod_time = datetime.fromtimestamp(os.path.getmtime(item_path)).strftime(
            "%Y-%m-%d %H:%M:%S"
        )

        file_list.append(
            {"name": item, "is_dir": is_dir, "size": size, "mod_time": mod_time}
        )

    return sorted(file_list, key=lambda x: (not x["is_dir"], x["name"].lower()))


@app.get("/datasites", response_class=HTMLResponse)
async def list_datasites(
    request: Request, server_settings: ServerSettings = Depends(get_server_settings)
):
    files = get_file_list(server_settings.snapshot_folder)
    template_path = current_dir / "templates" / "datasites.html"
    html = ""
    with open(template_path) as f:
        html = f.read()
    template = Template(html)

    html_content = template.render(
        {
            "request": request,
            "files": files,
            "current_path": "/",
        }
    )
    return html_content


@app.get("/datasites/{path:path}", response_class=HTMLResponse)
async def browse_datasite(
    request: Request,
    path: str,
    server_settings: ServerSettings = Depends(get_server_settings),
):
    if path == "":  # Check if path is empty (meaning "/datasites/")
        return RedirectResponse(url="/datasites")

    snapshot_folder = str(server_settings.snapshot_folder)
    datasite_part = path.split("/")[0]
    datasites = get_datasites(snapshot_folder)
    if datasite_part in datasites:
        slug = path[len(datasite_part) :]
        if slug == "":
            slug = "/"
        datasite_path = os.path.join(snapshot_folder, datasite_part)
        datasite_public = datasite_path + "/public"
        if not os.path.exists(datasite_public):
            return "No public datasite"

        slug_path = os.path.abspath(datasite_public + slug)
        if os.path.exists(slug_path) and os.path.isfile(slug_path):
            if slug_path.endswith(".html") or slug_path.endswith(".htm"):
                return FileResponse(slug_path)
            elif slug_path.endswith(".md"):
                with open(slug_path, "r") as file:
                    content = file.read()
                return PlainTextResponse(content)
            else:
                return FileResponse(slug_path, media_type="application/octet-stream")

        # show directory
        if not path.endswith("/"):
            return RedirectResponse(url=f"{path}/")

        index_file = os.path.abspath(slug_path + "/" + "index.html")
        if os.path.exists(index_file):
            with open(index_file, "r") as file:
                html_content = file.read()
            return HTMLResponse(content=html_content, status_code=200)

        if os.path.isdir(slug_path):
            files = get_file_list(slug_path)
            template_path = current_dir / "templates" / "folder.html"
            html = ""
            with open(template_path) as f:
                html = f.read()
            template = Template(html)
            html_content = template.render(
                {
                    "datasite": datasite_part,
                    "request": request,
                    "files": files,
                    "current_path": path,
                }
            )
            return html_content
        else:
            return f"Bad Slug {slug}"

    return f"No Datasite {datasite_part} exists"


@app.post("/register")
async def register(request: Request, users: Users = Depends(get_users)):
    data = await request.json()
    email = data["email"]
    token = users.create_user(email)
    logger.info(f"> {email} registering: {token}")
    return JSONResponse({"status": "success", "token": token}, status_code=200)


@app.post("/write", response_model=WriteResponse)
async def write(
    request: WriteRequest,
    server_settings: ServerSettings = Depends(get_server_settings),
) -> WriteResponse:
    try:
        email = request.email
        change = request.change
        reason = "ok"

        change.sync_folder = os.path.abspath(str(server_settings.snapshot_folder))
        result = True
        accepted = True
        if change.newer():
            if change.kind_write:
                if request.is_directory:
                    # Handle empty directory
                    os.makedirs(change.full_path, exist_ok=True)
                    result = True
                else:
                    bin_data = strtobin(request.data)
                    result = change.write(bin_data)
            elif change.kind_delete:
                if change.hash_equal_or_none():
                    result = change.delete()
                else:
                    reason = f"> 🔥 {change.kind} hash doesnt match so ignore {change}"
                    logger.info(reason)
                    accepted = False
            else:
                raise Exception(f"Unknown type of change kind. {change.kind}")
        else:
            reason = f"> 🔥 {change.kind} is older so ignore {change}"
            logger.info(reason)
            accepted = False

        if result:
            logger.info(f"> {email} {change.kind}: {change.internal_path}")
            return WriteResponse(
                status="success",
                change=change,
                accepted=accepted,
                reason=reason,
            )
        return WriteResponse(
            status="error",
            change=change,
            accepted=accepted,
            reason=reason,
        ), 400
    except Exception as e:
        logger.info("Exception writing", e)
        raise HTTPException(
            status_code=400,
            detail=f"Exception writing {e}",
        )


@app.post("/read", response_model=ReadResponse)
async def read(
    request: ReadRequest, server_settings: ServerSettings = Depends(get_server_settings)
) -> ReadResponse:
    email = request.email
    change = request.change
    change.sync_folder = os.path.abspath(str(server_settings.snapshot_folder))
    logger.info(f"> {email} {change.kind}: {change.internal_path}")
    # TODO: handle permissions, create and delete
    data = None
    if change.kind_write and not change.is_directory():
        data = bintostr(change.read())
    return ReadResponse(
        status="success",
        change=change.model_dump(mode="json"),
        data=data,
        is_directory=change.is_directory(),
    )


@app.post("/dir_state", response_model=DirStateResponse)
async def dir_state(
    request: DirStateRequest,
    server_settings: ServerSettings = Depends(get_server_settings),
) -> DirStateResponse:
    try:
        email = request.email
        sub_path = request.sub_path
        snapshot_folder = str(server_settings.snapshot_folder)
        full_path = os.path.join(snapshot_folder, sub_path)
        remote_dir_state = hash_dir(snapshot_folder, sub_path)

        # get the top level perm file
        try:
            perm_tree = PermissionTree.from_path(
                full_path, raise_on_corrupted_files=True
            )
        except Exception as e:
            print(f"Failed to parse permission tree: {full_path}")
            raise e

        # filter the read state for this user by the perm tree
        read_state = filter_read_state(email, remote_dir_state, perm_tree)
        remote_dir_state.tree = read_state

        if remote_dir_state:
            return DirStateResponse(
                sub_path=sub_path,
                dir_state=remote_dir_state,
                status="success",
            )
        raise HTTPException(status_code=400, detail={"status": "error"})
    except Exception as e:
        logger.exception("Failed to run /dir_state", e)


@app.get("/list_datasites", response_model=ListDatasitesResponse)
async def datasites(
    server_settings: ServerSettings = Depends(get_server_settings),
) -> ListDatasitesResponse:
    datasites = get_datasites(server_settings.snapshot_folder)
    if isinstance(datasites, list):
        return ListDatasitesResponse(
            datasites=datasites,
            status="success",
        )
    raise HTTPException(status_code=400, detail={"status": "error"})


@app.get("/install.sh")
async def install():
    install_script = current_dir / "templates" / "install.sh"
    return FileResponse(install_script, media_type="text/plain")


@app.get("/info")
async def info():
    return {
        "version": __version__,
    }


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Run FastAPI server")
    parser.add_argument(
        "--port",
        type=int,
        default=5001,
        help="Port to run the server on (default: 5001)",
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Run the server in debug mode with hot reloading",
    )
    parser.add_argument(
        "--ssl-keyfile",
        type=str,
        help="Path to SSL key file for HTTPS",
    )
    parser.add_argument(
        "--ssl-keyfile-password",
        type=str,
        help="SSL key file password for HTTPS",
    )
    parser.add_argument(
        "--ssl-certfile",
        type=str,
        help="Path to SSL certificate file for HTTPS",
    )

    args = parser.parse_args()
    return args


def main() -> None:
    args = parse_args()
    uvicorn_config = {
        "app": "syftbox.server.server:app" if args.debug else app,
        "host": "0.0.0.0",
        "port": args.port,
        "log_level": "debug" if args.debug else "info",
        "reload": args.debug,
    }

    uvicorn_config["ssl_keyfile"] = args.ssl_keyfile if args.ssl_keyfile else None
    uvicorn_config["ssl_certfile"] = args.ssl_certfile if args.ssl_certfile else None
    uvicorn_config["ssl_keyfile_password"] = (
        args.ssl_keyfile_password if args.ssl_keyfile_password else None
    )

    uvicorn.run(**uvicorn_config)


if __name__ == "__main__":
    main()



import json
import time

import pytest
from fastapi.testclient import TestClient

from syftbox.lib.lib import bintostr
from syftbox.server.server import app
from syftbox.server.settings import ServerSettings

TEST_DATASITE_NAME = "test_datasite@openmined.org"
TEST_FILE = "test_file.txt"
PERMFILE_FILE = "_.syftperm"
PERMFILE_DICT = {
    "admin": [TEST_DATASITE_NAME],
    "read": ["GLOBAL"],
    "write": [TEST_DATASITE_NAME],
}


@pytest.fixture(scope="function")
def client(monkeypatch, tmp_path):
    """Every client gets their own snapshot folder at `tmp_path`"""
    snapshot_folder = tmp_path / TEST_DATASITE_NAME
    settings = ServerSettings.from_data_folder(snapshot_folder)
    monkeypatch.setenv("SYFTBOX_DATA_FOLDER", str(settings.data_folder))
    monkeypatch.setenv("SYFTBOX_SNAPSHOT_FOLDER", str(settings.snapshot_folder))
    monkeypatch.setenv("SYFTBOX_USER_FILE_PATH", str(settings.user_file_path))

    datasite_name = TEST_DATASITE_NAME
    datasite = settings.snapshot_folder / datasite_name
    datasite.mkdir(parents=True)

    datafile = datasite / TEST_FILE
    datafile.touch()
    datafile.write_bytes(b"Hello, World!")

    permfile = datasite / PERMFILE_FILE
    permfile.touch()
    permfile.write_text(json.dumps(PERMFILE_DICT))

    with TestClient(app) as client:
        yield client


def test_register(client):
    data = {"email": "test@example.com"}
    response = client.post("/register", json=data)
    assert response.status_code == 200
    assert "token" in response.json()

    response = client.get("/list_datasites")
    assert response.status_code == 200


def test_write_file(client: TestClient):
    request_data = {
        "email": TEST_DATASITE_NAME,
        "change": {
            "kind": "write",
            "parent_path": TEST_DATASITE_NAME,
            "sub_path": "test_file.txt",
            "file_hash": "some_hash",
            "last_modified": time.time(),
        },
        "data": bintostr(b"Hello, World!"),
    }

    # Send POST request to /write endpoint
    response = client.post("/write", json=request_data)
    response.raise_for_status()
    data = response.json()
    print(data)


def test_list_datasites(client: TestClient):
    response = client.get("/list_datasites")
    assert response.status_code == 200

    assert len(response.json()["datasites"])

    response = client.get(f"/datasites/{TEST_DATASITE_NAME}/")
    assert response.status_code == 200


def test_read_file(client: TestClient):
    change = {
        "kind": "write",
        "parent_path": TEST_DATASITE_NAME,
        "sub_path": TEST_FILE,
        "file_hash": "some_hash",
        "last_modified": time.time(),
    }
    response = client.post(
        "/read", json={"email": TEST_DATASITE_NAME, "change": change}
    )

    response.raise_for_status()


def test_read_folder(client: TestClient):
    change = {
        "kind": "write",
        "parent_path": TEST_DATASITE_NAME,
        "sub_path": ".",
        "file_hash": "some_hash",
        "last_modified": time.time(),
    }
    response = client.post(
        "/read", json={"email": TEST_DATASITE_NAME, "change": change}
    )

    response.raise_for_status()


def test_dir_state(client: TestClient):
    response = client.post(
        "/dir_state", json={"email": TEST_DATASITE_NAME, "sub_path": "."}
    )

    response.raise_for_status()
    tree = response.json()["dir_state"]["tree"]
    assert "test_datasite@openmined.org/test_file.txt" in tree



"""
edge cases:
- if i locally modify something I dont have write access to, should sync revert my changes?
    - currently: it won't sync the changes to server, but it won't revert local changes
- why does delete take 5 seconds?
- what do we do when a permission file is corrupted?
"""

import json
import time
from collections.abc import Generator
from functools import partial
from pathlib import Path
from typing import Mapping, Union

import faker
import httpx
import pytest
from fastapi.testclient import TestClient

from syftbox.client.plugins.create_datasite import run as run_create_datasite_plugin
from syftbox.client.plugins.init import run as run_init_plugin
from syftbox.client.plugins.sync import do_sync
from syftbox.lib.lib import ClientConfig, SharedState, SyftPermission, perm_file_path
from syftbox.server.server import app as server_app
from syftbox.server.server import lifespan as server_lifespan
from syftbox.server.settings import ServerSettings

fake = faker.Faker()

DirTree = Mapping[str, Union[str, "DirTree"]]


def create_local_tree(base_path: Path, tree: DirTree) -> None:
    print(f"creating tree at {base_path}, {type(base_path)}")
    for name, content in tree.items():
        local_path = base_path / name

        if isinstance(content, str):
            local_path.write_text(content)
        elif isinstance(content, SyftPermission):
            content.save(path=str(local_path))
        elif isinstance(content, dict):
            local_path.mkdir(parents=True, exist_ok=True)
            create_local_tree(local_path, content)


@pytest.fixture(scope="function")
def datasite_1(tmp_path: Path, server_client: TestClient) -> ClientConfig:
    email = "user_1@openmined.org"
    return setup_datasite(tmp_path, server_client, email)


@pytest.fixture(scope="function")
def datasite_2(tmp_path: Path, server_client: TestClient) -> ClientConfig:
    email = "user_2@openmined.org"
    return setup_datasite(tmp_path, server_client, email)


def setup_datasite(
    tmp_path: Path, server_client: TestClient, email: str
) -> ClientConfig:
    client_path = tmp_path / email
    client_path.unlink(missing_ok=True)
    client_path.mkdir(parents=True)

    client_config = ClientConfig(
        config_path=str(client_path / "client_config.json"),
        sync_folder=str(client_path / "sync"),
        email=email,
        server_url=str(server_client.base_url),
        autorun_plugins=[],
    )

    client_config._server_client = server_client

    shared_state = SharedState(client_config=client_config)
    run_init_plugin(shared_state)
    run_create_datasite_plugin(shared_state)
    wait_for_datasite_setup(client_config)
    return client_config


@pytest.fixture(scope="function")
def server_client(tmp_path: Path) -> Generator[TestClient, None, None]:
    print("Using test dir", tmp_path)
    path = tmp_path / "server"
    path.mkdir()

    settings = ServerSettings.from_data_folder(path)
    lifespan_with_settings = partial(server_lifespan, settings=settings)
    server_app.router.lifespan_context = lifespan_with_settings

    with TestClient(server_app) as client:
        yield client


@pytest.fixture(scope="function")
def http_server_client():
    with httpx.Client(base_url="http://localhost:5001") as client:
        yield client


def wait_for_datasite_setup(client_config: ClientConfig, timeout=5):
    print("waiting for datasite setup...")

    perm_file = perm_file_path(str(client_config.datasite_path))

    t0 = time.time()
    while time.time() - t0 < timeout:
        perm_file_exists = Path(perm_file).exists()
        is_registered = client_config.is_registered
        if perm_file_exists and is_registered:
            print("Datasite setup complete")
            return
        time.sleep(1)

    raise TimeoutError("Datasite setup took too long")


def create_random_file(client_config: ClientConfig, sub_path: str = "") -> Path:
    relative_path = Path(sub_path) / fake.file_name(extension="json")
    file_path = client_config.datasite_path / relative_path
    content = {"body": fake.text()}
    file_path.write_text(json.dumps(content))

    path_in_datasite = file_path.relative_to(client_config.sync_folder)
    return path_in_datasite


def assert_files_not_on_datasite(datasite: ClientConfig, files: list[Path]):
    for file in files:
        assert not (
            datasite.sync_folder / file
        ).exists(), f"File {file} exists on datasite {datasite.email}"


def assert_files_on_datasite(datasite: ClientConfig, files: list[Path]):
    for file in files:
        assert (
            datasite.sync_folder / file
        ).exists(), f"File {file} does not exist on datasite {datasite.email}"


def assert_files_on_server(server_client: TestClient, files: list[Path]):
    server_settings: ServerSettings = server_client.app_state["server_settings"]
    for file in files:
        assert (
            server_settings.snapshot_folder / file
        ).exists(), f"File {file} does not exist on server"


def assert_dirtree_exists(base_path: Path, tree: DirTree) -> None:
    for name, content in tree.items():
        local_path = base_path / name

        if isinstance(content, str):
            assert local_path.read_text() == content
        elif isinstance(content, SyftPermission):
            assert json.loads(local_path.read_text()) == content.to_dict()
        elif isinstance(content, dict):
            assert local_path.is_dir()
            assert_dirtree_exists(local_path, content)


def test_create_public_file(
    server_client: TestClient, datasite_1: ClientConfig, datasite_2: ClientConfig
):
    # Two datasites create and sync a random file each

    datasite_1_shared_state = SharedState(client_config=datasite_1)
    datasite_2_shared_state = SharedState(client_config=datasite_2)

    file_path_1 = create_random_file(datasite_1, "public")
    file_path_2 = create_random_file(datasite_2, "public")
    assert_files_on_datasite(datasite_1, [file_path_1])
    assert_files_on_datasite(datasite_2, [file_path_2])

    # client 1 syncs
    do_sync(datasite_1_shared_state)
    assert_files_on_server(server_client, [file_path_1])
    assert_files_on_datasite(datasite_1, [file_path_1])

    # client 2 syncs
    do_sync(datasite_2_shared_state)
    assert_files_on_server(server_client, [file_path_1, file_path_2])
    assert_files_on_datasite(datasite_1, [file_path_1])
    assert_files_on_datasite(datasite_2, [file_path_1, file_path_2])

    # client 1 syncs again
    do_sync(datasite_1_shared_state)
    assert_files_on_server(server_client, [file_path_1, file_path_2])
    assert_files_on_datasite(datasite_1, [file_path_1, file_path_2])


def test_modify_public_file(
    server_client: TestClient, datasite_1: ClientConfig, datasite_2: ClientConfig
):
    # Two datasites create and sync a random file each

    datasite_1_shared_state = SharedState(client_config=datasite_1)
    datasite_2_shared_state = SharedState(client_config=datasite_2)

    file_path_1 = create_random_file(datasite_1, "public")
    assert_files_on_datasite(datasite_1, [file_path_1])

    # client 1 syncs
    do_sync(datasite_1_shared_state)
    assert_files_on_server(server_client, [file_path_1])

    # client 2 syncs
    do_sync(datasite_2_shared_state)
    assert_files_on_datasite(datasite_2, [file_path_1])

    # client 1 modifies
    (datasite_1.sync_folder / file_path_1).write_text("modified")
    do_sync(datasite_1_shared_state)

    # client 2 gets the modification
    do_sync(datasite_2_shared_state)
    assert (datasite_2.sync_folder / file_path_1).read_text() == "modified"


def test_delete_public_file(
    server_client: TestClient, datasite_1: ClientConfig, datasite_2: ClientConfig
):
    # Two datasites create and sync a random file each
    datasite_1_shared_state = SharedState(client_config=datasite_1)
    datasite_2_shared_state = SharedState(client_config=datasite_2)

    file_path_1 = create_random_file(datasite_1, "public")
    assert_files_on_datasite(datasite_1, [file_path_1])

    # client 1 syncs
    do_sync(datasite_1_shared_state)
    assert_files_on_server(server_client, [file_path_1])

    # client 2 syncs
    do_sync(datasite_2_shared_state)
    assert_files_on_datasite(datasite_2, [file_path_1])

    # client 1 deletes
    (datasite_1.sync_folder / file_path_1).unlink()

    # deletion is only synced after a few seconds, so first sync does not delete
    do_sync(datasite_1_shared_state)
    do_sync(datasite_2_shared_state)
    assert_files_on_datasite(datasite_2, [file_path_1])

    # after a few seconds the file is deleted
    time.sleep(5)
    do_sync(datasite_1_shared_state)
    do_sync(datasite_2_shared_state)
    assert_files_on_datasite(datasite_2, [file_path_1])


def test_move_file(
    server_client: TestClient, datasite_1: ClientConfig, datasite_2: ClientConfig
):
    datasite_1_shared_state = SharedState(client_config=datasite_1)
    datasite_2_shared_state = SharedState(client_config=datasite_2)

    tree = {
        "folder1": {
            "_.syftperm": SyftPermission.mine_with_public_read(datasite_1.email),
            "file1.txt": "content1",
        },
    }

    create_local_tree(Path(datasite_1.datasite_path), tree)

    do_sync(datasite_1_shared_state)
    do_sync(datasite_2_shared_state)

    assert_dirtree_exists(Path(datasite_2.sync_folder) / datasite_1.email, tree)

    # move file1 to new folder
    file1 = Path(datasite_1.datasite_path) / "folder1" / "file1.txt"

    new_tree = {
        "folder2": {
            "_.syftperm": SyftPermission.mine_with_public_read(datasite_1.email),
        },
    }

    do_sync(datasite_1_shared_state)
    do_sync(datasite_2_shared_state)

    # wait for delete
    time.sleep(5)

    create_local_tree(Path(datasite_1.datasite_path), new_tree)
    file1.rename(Path(datasite_1.datasite_path) / "folder2" / "file1.txt")

    do_sync(datasite_1_shared_state)
    do_sync(datasite_2_shared_state)

    assert_files_not_on_datasite(
        datasite_2, [Path(datasite_1.email) / "folder1" / "file1.txt"]
    )
    assert_dirtree_exists(Path(datasite_2.sync_folder) / datasite_1.email, new_tree)


def test_sync_with_permissions(
    server_client: TestClient, datasite_1: ClientConfig, datasite_2: ClientConfig
):
    # TODO split in multiple tests
    datasite_1_shared_state = SharedState(client_config=datasite_1)
    datasite_2_shared_state = SharedState(client_config=datasite_2)

    snapshot_folder = server_client.app_state["server_settings"].snapshot_folder
    snapshot_datasite_1 = snapshot_folder / datasite_1.email

    do_sync(datasite_1_shared_state)
    do_sync(datasite_2_shared_state)

    tree = {
        "public_read": {
            "_.syftperm": SyftPermission.mine_with_public_read(datasite_1.email),
            "public_read.json": "content1",
        },
        "private": {
            "_.syftperm": SyftPermission.mine_no_permission(datasite_1.email),
            "private.json": "content2",
        },
        "public_write": {
            "_.syftperm": SyftPermission.mine_with_public_write(datasite_1.email),
            "public_write.json": "content3",
        },
        "no_permission": {
            "no_permission.json": "content4",
        },
    }

    # Trees filtered by permission for datasite 2
    public_read_folders = ["public_read", "public_write"]
    public_read_tree = {k: v for k, v in tree.items() if k in public_read_folders}

    create_local_tree(Path(datasite_1.datasite_path), tree)
    assert_dirtree_exists(Path(datasite_1.datasite_path), tree)

    do_sync(datasite_1_shared_state)
    assert_dirtree_exists(snapshot_datasite_1, tree)

    do_sync(datasite_2_shared_state)
    # public files exist, private files do not
    assert_dirtree_exists(
        Path(datasite_2.sync_folder) / datasite_1.email, public_read_tree
    )
    assert_files_not_on_datasite(
        datasite_2,
        [
            Path(datasite_1.email) / "private" / "private.json",
            Path(datasite_1.email) / "no_permission" / "no_permission.json",
        ],
    )

    # DS2 writes to public write file
    public_write_file = (
        Path(datasite_2.sync_folder)
        / datasite_1.email
        / "public_write"
        / "public_write.json"
    )
    public_write_file.write_text("modified")

    do_sync(datasite_2_shared_state)
    do_sync(datasite_1_shared_state)

    assert (
        Path(datasite_1.datasite_path) / "public_write" / "public_write.json"
    ).read_text() == "modified"

    # DS2 writes to public read file, should not be synced
    public_read_file = (
        Path(datasite_2.sync_folder)
        / datasite_1.email
        / "public_read"
        / "public_read.json"
    )
    public_read_file.write_text("modified")

    do_sync(datasite_2_shared_state)
    do_sync(datasite_1_shared_state)

    # Server and datasite 1 should not have the modification
    assert (
        snapshot_datasite_1 / "public_read" / "public_read.json"
    ).read_text() == "content1"
    assert (
        Path(datasite_1.datasite_path) / "public_read" / "public_read.json"
    ).read_text() == "content1"


def test_corrupted_permissions(
    server_client: TestClient, datasite_1: ClientConfig, datasite_2: ClientConfig
):
    datasite_1_shared_state = SharedState(client_config=datasite_1)
    datasite_2_shared_state = SharedState(client_config=datasite_2)

    do_sync(datasite_1_shared_state)
    do_sync(datasite_2_shared_state)

    dir_tree = {
        "corrupted_folder": {
            "_.syftperm": SyftPermission.mine_with_public_read(datasite_1.email),
            "file.txt": "content",
            "corrupted_subfolder": {
                "_.syftperm": SyftPermission.mine_with_public_read(datasite_1.email),
                "file.txt": "content",
            },
        },
        "normal_folder": {
            "_.syftperm": SyftPermission.mine_with_public_read(datasite_1.email),
            "file.txt": "content",
        },
    }

    create_local_tree(Path(datasite_1.datasite_path), dir_tree)

    do_sync(datasite_1_shared_state)
    do_sync(datasite_2_shared_state)

    # Corrupt the permission file and do an update
    permission_file = Path(datasite_1.datasite_path) / "corrupted_folder" / "_.syftperm"
    permission_file.write_text("corrupted")

    # Make local changes
    # expected behaviour: files under corrupted permissions are not synced, other files are
    corrupted_file_to_update = (
        Path(datasite_1.datasite_path) / "corrupted_folder" / "file.txt"
    )
    corrupted_subfolder_file_to_update = (
        Path(datasite_1.datasite_path)
        / "corrupted_folder"
        / "corrupted_subfolder"
        / "file.txt"
    )
    normal_file_to_update = (
        Path(datasite_1.datasite_path) / "normal_folder" / "file.txt"
    )
    corrupted_file_to_update.write_text("updated")
    corrupted_subfolder_file_to_update.write_text("updated")
    normal_file_to_update.write_text("updated")

    # NOTE fix this, need to sleep in order to detect faulty deletes
    time.sleep(5)

    do_sync(datasite_1_shared_state)
    do_sync(datasite_2_shared_state)

    # Corrupted folders and subfolders are not synced
    assert (
        Path(datasite_2.sync_folder)
        / datasite_1.email
        / "corrupted_folder"
        / "file.txt"
    ).read_text() == "content"
    assert (
        Path(datasite_2.sync_folder)
        / datasite_1.email
        / "corrupted_folder"
        / "corrupted_subfolder"
        / "file.txt"
    ).read_text() == "content"

    # Normal folder is synced
    assert (
        Path(datasite_2.sync_folder) / datasite_1.email / "normal_folder" / "file.txt"
    ).read_text() == "updated"

    # Fix the corrupted permission file
    SyftPermission.mine_with_public_read(datasite_1.email).save(
        path=str(permission_file)
    )

    do_sync(datasite_1_shared_state)
    do_sync(datasite_2_shared_state)

    # Corrupted folders and subfolders are now synced
    assert (
        Path(datasite_2.sync_folder)
        / datasite_1.email
        / "corrupted_folder"
        / "file.txt"
    ).read_text() == "updated"
    assert (
        Path(datasite_2.sync_folder)
        / datasite_1.email
        / "corrupted_folder"
        / "corrupted_subfolder"
        / "file.txt"
    ).read_text() == "updated"



from __future__ import annotations

import base64
import hashlib
import json
import os
import platform
import re
import threading
import zlib
from dataclasses import dataclass, field
from datetime import datetime, timezone
from pathlib import Path
from threading import Lock

import httpx
import requests
from loguru import logger
from typing_extensions import Any, Optional, Self

from syftbox.client.utils import macos
from syftbox.server.models import (
    DirState,
    FileInfo,
    get_file_hash,
    get_file_last_modified,
)

from .exceptions import ClientConfigException

current_dir = Path(__file__).parent
ASSETS_FOLDER = current_dir.parent / "assets"
DEFAULT_PORT = 8082
ICON_FOLDER = ASSETS_FOLDER / "icon"
DEFAULT_SYNC_FOLDER = os.path.expanduser("~/Desktop/SyftBox")
DEFAULT_CONFIG_FOLDER = os.path.expanduser("~/.syftbox")
DEFAULT_CONFIG_PATH = os.path.join(DEFAULT_CONFIG_FOLDER, "client_config.json")
DEFAULT_LOGS_PATH = os.path.join(DEFAULT_CONFIG_FOLDER, "logs", "syftbox.log")

USER_GROUP_GLOBAL = "GLOBAL"

ICON_FILE = "Icon"  # special
IGNORE_FILES = []


def perm_file_path(path: str) -> str:
    return f"{path}/_.syftperm"


def is_primitive_json_serializable(obj):
    if isinstance(obj, (str, int, float, bool, type(None))):
        return True
    return False


def pack(obj) -> Any:
    if is_primitive_json_serializable(obj):
        return obj

    if hasattr(obj, "to_dict"):
        return obj.to_dict()

    if isinstance(obj, list):
        return [pack(val) for val in obj]

    if isinstance(obj, dict):
        return {k: pack(v) for k, v in obj.items()}

    raise Exception(f"Unable to pack type: {type(obj)} value: {obj}")


class Jsonable:
    def to_dict(self) -> dict:
        output = {}
        for k, v in self.__dict__.items():
            if k.startswith("_"):
                continue
            output[k] = pack(v)
        return output

    def __iter__(self):
        for key, val in self.to_dict().items():
            if key.startswith("_"):
                yield key, val

    def __getitem__(self, key):
        if key.startswith("_"):
            return None
        return self.to_dict()[key]

    @classmethod
    def load(cls, filepath: str) -> Self:
        try:
            with open(filepath) as f:
                data = f.read()
                d = json.loads(data)
                return cls(**d)
        except Exception as e:
            raise e
            logger.info(f"Unable to load jsonable file: {filepath}. {e}")
        return None

    def save(self, filepath: str) -> None:
        d = self.to_dict()
        with open(Path(filepath).expanduser(), "w") as f:
            f.write(json.dumps(d))


@dataclass
class SyftPermission(Jsonable):
    admin: list[str]
    read: list[str]
    write: list[str]
    filepath: str | None = None
    terminal: bool = False

    @classmethod
    def datasite_default(cls, email: str) -> Self:
        return cls(
            admin=[email],
            read=[email],
            write=[email],
        )

    def has_read_permission(self, email: str) -> bool:
        return email in self.read or USER_GROUP_GLOBAL in self.read

    def __eq__(self, other):
        if not isinstance(other, SyftPermission):
            return NotImplemented
        return (
            self.admin == other.admin
            and self.read == other.read
            and self.write == other.write
            and self.filepath == other.filepath
            and self.terminal == other.terminal
        )

    def perm_path(self, path=None) -> str:
        if path is not None:
            self.filepath = path

        if self.filepath is None:
            raise Exception(f"Saving requites a path: {self}")

        if os.path.isdir(self.filepath):
            self.filepath = perm_file_path(self.filepath)
        return self.filepath

    def save(self, path=None) -> bool:
        self.perm_path(path=path)
        if self.filepath.endswith(".syftperm"):
            super().save(self.filepath)
        else:
            raise Exception(f"Perm file must end in .syftperm. {self.filepath}")
        return True

    def ensure(self, path=None) -> bool:
        # make sure the contents matches otherwise write it
        self.perm_path(path=path)
        try:
            prev_perm_file = SyftPermission.load(self.filepath)
            if self == prev_perm_file:
                # no need to write
                return True
        except Exception:
            pass
        return self.save(path)

    @classmethod
    def no_permission(cls) -> Self:
        return cls(admin=[], read=[], write=[])

    @classmethod
    def mine_no_permission(cls, email: str) -> Self:
        return cls(admin=[email], read=[], write=[])

    @classmethod
    def mine_with_public_read(cls, email: str) -> Self:
        return cls(admin=[email], read=[email, "GLOBAL"], write=[email])

    @classmethod
    def mine_with_public_write(cls, email: str) -> Self:
        return cls(admin=[email], read=[email, "GLOBAL"], write=[email, "GLOBAL"])

    @classmethod
    def theirs_with_my_read(cls, their_email, my_email: str) -> Self:
        return cls(
            admin=[their_email], read=[their_email, my_email], write=[their_email]
        )

    @classmethod
    def theirs_with_my_read_write(cls, their_email, my_email: str) -> Self:
        return cls(
            admin=[their_email],
            read=[their_email, my_email],
            write=[their_email, my_email],
        )

    def __repr__(self) -> str:
        string = "SyftPermission:\n"
        string += f"{self.filepath}\n"
        string += "ADMIN: ["
        for v in self.admin:
            string += v + ", "
        string += "]\n"

        string += "READ: ["
        for r in self.read:
            string += r + ", "
        string += "]\n"

        string += "WRITE: ["
        for w in self.write:
            string += w + ", "
        string += "]\n"
        return string


def bintostr(binary_data):
    return base64.b85encode(zlib.compress(binary_data)).decode("utf-8")


def strtobin(encoded_data):
    return zlib.decompress(base64.b85decode(encoded_data.encode("utf-8")))


def get_symlink(file_path) -> str:
    return os.readlink(file_path)


def is_symlink(file_path) -> bool:
    return os.path.islink(file_path)


# def symlink_to_syftlink(file_path):
#     return SyftLink.from_path(file_path)


# def convert_to_symlink(path):
#     if not is_symlink(path):
#         raise Exception(f"Cant convert a non symlink {path}")
#     abs_path = get_symlink(path)
#     syft_link = symlink_to_syftlink(abs_path)
#     return str(syft_link)


def ignore_dirs(directory: str, root: str, ignore_folders=None) -> bool:
    if ignore_folders is not None:
        for ignore_folder in ignore_folders:
            if root.endswith(ignore_folder):
                return True
    return False


def hash_dir(
    sync_folder: str,
    sub_path: str,
    ignore_folders: list | None = None,
) -> DirState:
    state_dict = {}
    full_path = os.path.join(sync_folder, sub_path)
    for root, dirs, files in os.walk(full_path):
        if not ignore_dirs(full_path, root, ignore_folders):
            for file in files:
                if not ignore_file(full_path, root, file):
                    path = os.path.join(root, file)
                    rel_path = os.path.relpath(path, full_path)
                    file_info = FileInfo(
                        file_hash=get_file_hash(path),
                        last_modified=get_file_last_modified(path),
                    )
                    state_dict[rel_path] = file_info

    utc_unix_timestamp = datetime.now(timezone.utc).timestamp()
    dir_state = DirState(
        tree=state_dict,
        timestamp=utc_unix_timestamp,
        sync_folder=sync_folder,
        sub_path=sub_path,
    )
    return dir_state


def ignore_file(directory: str, root: str, filename: str) -> bool:
    if directory == root:
        if filename.startswith(ICON_FILE):
            return True
        if filename in IGNORE_FILES:
            return True
    if filename == ".DS_Store":
        return True
    return False


def get_datasites(sync_folder: str | Path) -> list[str]:
    sync_folder = (
        str(sync_folder.resolve()) if isinstance(sync_folder, Path) else sync_folder
    )
    datasites = []
    folders = os.listdir(sync_folder)
    for folder in folders:
        if "@" in folder:
            datasites.append(folder)
    return datasites


def build_tree_string(paths_dict, prefix=""):
    lines = []
    items = list(paths_dict.items())

    for index, (key, value) in enumerate(items):
        # Determine if it's the last item in the current directory level
        connector = "└── " if index == len(items) - 1 else "├── "
        lines.append(f"{prefix}{connector}{repr(key)}")

        # Prepare the prefix for the next level
        if isinstance(value, dict):
            extension = "    " if index == len(items) - 1 else "│   "
            lines.append(build_tree_string(value, prefix + extension))

    return "\n".join(lines)


@dataclass
class PermissionTree(Jsonable):
    tree: dict[str, SyftPermission]
    parent_path: str
    root_perm: SyftPermission | None

    corrupted_permission_files: list[str] = field(default_factory=list)

    @classmethod
    def from_path(cls, parent_path, raise_on_corrupted_files: bool = False) -> Self:
        corrupted_permission_files = []
        perm_dict = {}
        for root, dirs, files in os.walk(parent_path):
            for file in files:
                if file.endswith(".syftperm"):
                    path = os.path.join(root, file)
                    try:
                        perm_dict[path] = SyftPermission.load(path)
                    except Exception:
                        corrupted_permission_files.append(path)

        root_perm = None
        root_perm_path = perm_file_path(parent_path)
        if root_perm_path in perm_dict:
            root_perm = perm_dict[root_perm_path]

        if corrupted_permission_files:
            if raise_on_corrupted_files:
                raise ValueError(
                    f"Found corrupted permission files: {corrupted_permission_files}"
                )
            logger.warning(
                f"Found corrupted permission files: {corrupted_permission_files}"
            )

        return cls(
            root_perm=root_perm,
            tree=perm_dict,
            parent_path=parent_path,
            corrupted_permission_files=corrupted_permission_files,
        )

    def has_corrupted_permission(self, path: str | Path) -> bool:
        path = Path(path).resolve()
        corrupted_permission_paths = [
            Path(p).parent.resolve() for p in self.corrupted_permission_files
        ]
        for perm_path in corrupted_permission_paths:
            if path.is_relative_to(perm_path):
                return True
        return False

    @property
    def root_or_default(self) -> SyftPermission:
        if self.root_perm:
            return self.root_perm
        return SyftPermission.no_permission()

    def permission_for_path(self, path: str) -> SyftPermission:
        parent_path = os.path.normpath(self.parent_path)
        current_perm = self.root_or_default

        # default
        if parent_path not in path:
            return current_perm

        sub_path = path.replace(parent_path, "")
        current_perm_level = parent_path
        for part in sub_path.split("/"):
            if part == "":
                continue

            current_perm_level += "/" + part
            next_perm_file = perm_file_path(current_perm_level)
            if next_perm_file in self.tree:
                # we could do some overlay with defaults but
                # for now lets just use a fully defined overwriting perm file
                next_perm = self.tree[next_perm_file]
                current_perm = next_perm

            if current_perm.terminal:
                return current_perm

        return current_perm

    def __repr__(self) -> str:
        return f"PermissionTree: {self.parent_path}\n" + build_tree_string(self.tree)


def filter_read_state(user_email: str, dir_state: DirState, perm_tree: PermissionTree):
    filtered_tree = {}
    root_dir = dir_state.sync_folder + "/" + dir_state.sub_path
    for file_path, file_info in dir_state.tree.items():
        full_path = root_dir + "/" + file_path
        perm_file_at_path = perm_tree.permission_for_path(full_path)
        if (
            user_email in perm_file_at_path.read
            or "GLOBAL" in perm_file_at_path.read
            or user_email in perm_file_at_path.admin
        ):
            filtered_tree[file_path] = file_info
    return filtered_tree


class ResettableTimer:
    def __init__(self, timeout, callback, *args, **kwargs):
        self.timeout = timeout
        self.callback = callback
        self.args = args
        self.kwargs = kwargs
        self.timer = None
        self.lock = threading.Lock()

    def _run_callback(self):
        with self.lock:
            self.timer = None
        self.callback(*self.args, **self.kwargs)

    def start(self, *args, **kwargs):
        with self.lock:
            if self.timer:
                self.timer.cancel()

            # If new arguments are passed in start, they will overwrite the initial ones
            if args or kwargs:
                self.args = args
                self.kwargs = kwargs

            self.timer = threading.Timer(self.timeout, self._run_callback)
            self.timer.start()

    def cancel(self):
        with self.lock:
            if self.timer:
                self.timer.cancel()
                self.timer = None


class SharedState:
    def __init__(self, client_config: ClientConfig):
        self.data = {}
        self.lock = Lock()
        self.client_config = client_config
        self.timers: dict[str:ResettableTimer] = {}
        self.fs_events = []

    @property
    def sync_folder(self) -> str:
        return self.client_config.sync_folder

    def get(self, key, default=None):
        with self.lock:
            if key == "my_datasites":
                return self._get_datasites()
            return self.data.get(key, default)

    def set(self, key, value):
        with self.lock:
            self.data[key] = value

    def _get_datasites(self):
        syft_folder = self.data.get(self.client_config.sync_folder)
        if not syft_folder or not os.path.exists(syft_folder):
            return []

        return [
            folder
            for folder in os.listdir(syft_folder)
            if os.path.isdir(os.path.join(syft_folder, folder))
        ]


def get_root_data_path() -> Path:
    # get the PySyft / data directory to share datasets between notebooks
    # on Linux and MacOS the directory is: ~/.syft/data"
    # on Windows the directory is: C:/Users/$USER/.syft/data

    data_dir = Path.home() / ".syft" / "data"
    data_dir.mkdir(parents=True, exist_ok=True)

    return data_dir


def autocache(
    url: str, extension: str | None = None, cache: bool = True
) -> Path | None:
    try:
        data_path = get_root_data_path()
        file_hash = hashlib.sha256(url.encode("utf8")).hexdigest()
        filename = file_hash
        if extension:
            filename += f".{extension}"
        file_path = data_path / filename
        if os.path.exists(file_path) and cache:
            return file_path
        return download_file(url, file_path)
    except Exception as e:
        logger.info(f"Failed to autocache: {url}. {e}")
        return None


def download_file(url: str, full_path: str | Path) -> Path | None:
    full_path = Path(full_path)
    if not full_path.exists():
        r = requests.get(url, allow_redirects=True, verify=verify_tls())  # nosec
        if not r.ok:
            logger.info(f"Got {r.status_code} trying to download {url}")
            return None
        full_path.parent.mkdir(parents=True, exist_ok=True)
        full_path.write_bytes(r.content)
    return full_path


def verify_tls() -> bool:
    return not str_to_bool(str(os.environ.get("IGNORE_TLS_ERRORS", "0")))


def str_to_bool(bool_str: str | None) -> bool:
    result = False
    bool_str = str(bool_str).lower()
    if bool_str == "true" or bool_str == "1":
        result = True
    return result


def validate_email(email: str) -> bool:
    # Define a regex pattern for a valid email
    email_regex = r"^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$"

    # Use the match method to check if the email fits the pattern
    if re.match(email_regex, email):
        return True
    return False


@dataclass
class Client(Jsonable):
    config_path: Path
    sync_folder: Path | None = None
    port: int | None = None
    email: str | None = None
    token: int | None = None
    server_url: str = "http://localhost:5001"
    email_token: str | None = None
    autorun_plugins: list[str] | None = field(
        default_factory=lambda: ["init", "create_datasite", "sync", "apps"]
    )
    _server_client: httpx.Client | None = None

    @property
    def is_registered(self) -> bool:
        return self.token is not None

    @property
    def server_client(self) -> httpx.Client:
        if self._server_client is None:
            self._server_client = httpx.Client(
                base_url=self.server_url,
                follow_redirects=True,
            )
        return self._server_client

    def close(self):
        if self._server_client:
            self._server_client.close()

    def save(self, path: str | None = None) -> None:
        if path is None:
            path = self.config_path
        super().save(path)

    @property
    def datasite_path(self) -> Path:
        return os.path.join(self.sync_folder, self.email)

    @property
    def manifest_path(self) -> Path:
        return os.path.join(self.datasite_path, "public/manifest/manifest.json")

    def get_datasites(self: str) -> list[str]:
        datasites = []
        folders = os.listdir(self.sync_folder)
        for folder in folders:
            if "@" in folder:
                datasites.append(folder)
        return datasites

    def use(self):
        os.environ["SYFTBOX_CURRENT_CLIENT"] = self.config_path
        os.environ["SYFTBOX_SYNC_DIR"] = self.sync_folder
        logger.info(f"> Setting Sync Dir to: {self.sync_folder}")

    def create_folder(self, path: str, permission: SyftPermission):
        os.makedirs(path, exist_ok=True)
        permission.save(path)

    @property
    def root_dir(self) -> Path:
        root_dir = Path(os.path.abspath(os.path.dirname(self.file_path) + "/../"))
        return root_dir

    def create_public_folder(self, path: str):
        full_path = self.root_dir / path
        os.makedirs(str(full_path), exist_ok=True)
        public_read = SyftPermission.mine_with_public_read(email=self.datasite)
        public_read.save(full_path)
        return Path(full_path)

    @classmethod
    def load(cls, filepath: str | None = None) -> Self:
        try:
            if filepath is None:
                config_path = os.getenv(
                    "SYFTBOX_CLIENT_CONFIG_PATH", DEFAULT_CONFIG_PATH
                )
                filepath = config_path
            return super().load(filepath)
        except Exception:
            raise ClientConfigException(
                f"Unable to load Client config from {filepath}."
                "If you are running this outside of syftbox app runner you must supply "
                "the Client config path like so: \n"
                "SYFTBOX_CLIENT_CONFIG_PATH=~/.syftbox/client_config.json"
            )


ClientConfig = Client


def get_user_input(prompt, default: Optional[str] = None):
    if default:
        prompt = f"{prompt} (default: {default}): "
    user_input = input(prompt).strip()
    return user_input if user_input else default


def load_or_create_config(args) -> ClientConfig:
    syft_config_dir = os.path.abspath(os.path.expanduser("~/.syftbox"))
    os.makedirs(syft_config_dir, exist_ok=True)

    client_config = None
    try:
        client_config = ClientConfig.load(args.config_path)
    except Exception:
        pass

    if client_config is None and args.config_path:
        config_path = os.path.abspath(os.path.expanduser(args.config_path))
        client_config = ClientConfig(config_path=config_path)

    if client_config is None:
        # config_path = get_user_input("Path to config file?", DEFAULT_CONFIG_PATH)
        config_path = os.path.abspath(os.path.expanduser(config_path))
        client_config = ClientConfig(config_path=config_path)

    if args.sync_folder:
        sync_folder = os.path.abspath(os.path.expanduser(args.sync_folder))
        client_config.sync_folder = sync_folder

    if client_config.sync_folder is None:
        sync_folder = get_user_input(
            "Where do you want to Sync SyftBox to?",
            DEFAULT_SYNC_FOLDER,
        )
        sync_folder = os.path.abspath(os.path.expanduser(sync_folder))
        client_config.sync_folder = sync_folder

    if args.server:
        client_config.server_url = args.server

    if not os.path.exists(client_config.sync_folder):
        os.makedirs(client_config.sync_folder, exist_ok=True)

    if platform.system() == "Darwin":
        macos.copy_icon_file(ICON_FOLDER, client_config.sync_folder)

    if args.email:
        client_config.email = args.email

    if client_config.email is None:
        email = get_user_input("What is your email address? ")
        if not validate_email(email):
            raise Exception(f"Invalid email: {email}")
        client_config.email = email

    if args.port:
        client_config.port = args.port

    if client_config.port is None:
        port = int(get_user_input("Enter the port to use", DEFAULT_PORT))
        client_config.port = port

    email_token = os.environ.get("EMAIL_TOKEN", None)
    if email_token:
        client_config.email_token = email_token

    # Migrate Old Server URL to HTTPS
    if client_config.server_url == "http://20.168.10.234:8080":
        client_config.server_url = "https://syftbox.openmined.org"

    client_config.save(args.config_path)
    return client_config
